from sympy import *

nd, nx = symbols(r'nd, nx', integer=True)
kd = Idx(r'kd', nd)
kx = Idx(r'kx', nx)
xvec = IndexedBase(r'xvec', shape=(nd, nx))
x0 = IndexedBase(r'x0', shape=(nd))
out = IndexedBase(r'out', shape=(nx))

#print(get_indices(xvec[kd, kx] + xvec[kd, kx]))
#expr = Sum(xvec[kd, kx]*xvec[kd, kx], (kd, 0, nd-1))
expr = Sum(xvec[kd, kx]*x0[kd], (kd, 0, nd-1))

summed_indices = expr.variables
free_symbols = expr.expr_free_symbols

free_indices = set()

for s in free_symbols:
    if s.is_Indexed:
        for ind in s.indices:
            if not ind in summed_indices:
                free_indices.add(ind)

print(get_indices(expr))



"""
Created on Wed Aug 28 01:47:54 2020
@author: manal khallaayoune
"""

import numpy as np
import GPy
from GPy.kern.src.stationary import Stationary
from cosine_prod import Cosine_prod
import func_cosine_prod
import matplotlib.pyplot as plt

def test_kern_Cosine_prod():

    # Training points for 1D:
    x0train = np.linspace(-5,5,100).reshape(-1,1)
    x1train = np.linspace(-2,2,100).reshape(-1,1)
    x2train = np.linspace(0,9,100).reshape(-1,1)
    x3train = np.linspace(8,18,100).reshape(-1,1)
    
    #Training points for 2D:
    xtrain = np.hstack((x0train, x1train))
    
    ##Training points for 3D:
    #xtrain = np.hstack((x0train, x1train, x2train))
    
    ##Training points for 4D:
    #xtrain = np.hstack((x0train, x1train, x2train, x3train))
    
    
    ## 1D kernel:
    #k0 = Cosine_prod(1)
    #k0_K = k0.K(x0train,x1train)
    #dk0 = k0.dK_dX(x0train, x1train, 0)
    #dk0_2 = k0.dK2_dXdX2(x0train, x1train, 0, 0)
    
    # # 1D test:
#     l = np.array([1, 1])
    
#     K = np.zeros((len(x0train),len(x1train)))
#     for i in range(K.shape[0]):
#         for j in range(K.shape[1]):
#             K[i,j] = func.f_kern(xtrain[i,0], xtrain[i,1], xtrain[j,1], xtrain[j,1], l)
    
#     dK = np.zeros((len(x0train),len(x0train)))
#     for i in range(dK.shape[0]):
#         for j in range(dK.shape[1]):
#             # if dk0 = k0.dK_dX(x0train, x1train, 0) :
#             dK[i,j] = func.dkdx(xtrain[i,0], xtrain[i,1], xtrain[j,1], xtrain[j,1], l)
    
#     dK2 = np.zeros((len(x0train),len(x0train)))
#     for i in range(dK2.shape[0]):
#         for j in range(dK2.shape[1]):
#             # if dk0_2 = k0.dK2_dXdX2(x0train, x1train, 0, 0) :
#             dK2[i,j] = func.d2kdxdx0(xtrain[i,0], xtrain[i,1], xtrain[j,1], xtrain[j,1], l)
    
#     # 1D plot :
#     # plot the matrices
#     plt.figure()
#     plt.imshow(K)
#     plt.title('K sympy' '\n' '1D')
#     plt.figure()
#     plt.imshow(k0_K)
#     plt.title('K GPy' '\n' '1D')
#     # plot one of the courbes
#     plt.figure()
#     plt.plot(K[49,:])
#     plt.title('K sympy' '\n' '1D')
#     plt.figure()
#     plt.plot(k0_K[49,:])
#     plt.title('K GPy' '\n' '1D')
    
#     plt.figure()
#     plt.imshow(dK)
#     plt.title('1st derivative sympy' '\n' '1D')
#     plt.figure()
#     plt.imshow(dk0)
#     plt.title('1st derivative GPy' '\n' '1D')
#     plt.figure()
#     plt.plot(dK[49,:])
#     plt.title('1st derivative sympy' '\n' '1D')
#     plt.figure()
#     plt.plot(dk0[49,:])
#     plt.title('1st derivative GPy' '\n' '1D')
    
#     plt.figure()
#     plt.imshow(dK2)
#     plt.title('2nd derivative sympy' '\n' '1D')
#     plt.figure()
#     plt.imshow(dk0_2)
#     plt.title('2nd derivative GPy' '\n' '1D')
#     plt.figure()
#     plt.plot(dK2[49,:])
#     plt.title('2nd derivative sympy' '\n' '1D')
#     plt.figure()
#     plt.plot(dk0_2[49,:])
#     plt.title('2nd derivative GPy' '\n' '1D')


    # 2D kernel:
    k0 = Cosine_prod(2)
    k0_K = k0.K(xtrain,xtrain)
    dk0 = k0.dK_dX(xtrain, xtrain, 0)
    dk0_2 = k0.dK2_dXdX2(xtrain, xtrain, 0, 0)
    
    # 2D test:
    l = np.array([1, 1])
    
    K = np.zeros((len(x0train),len(x0train)))
    for i in range(K.shape[0]):
        for j in range(K.shape[1]):
            K[i,j] = func.f_kern(xtrain[i,0], xtrain[i,1], xtrain[j,0], xtrain[j,1], l)
    
    dK = np.zeros((len(x0train),len(x0train)))
    for i in range(dK.shape[0]):
        for j in range(dK.shape[1]):
            # if dk0 = k0.dK_dX(xtrain, xtrain, 0) :
            dK[i,j] = func.dkdx(xtrain[i,0], xtrain[i,1], xtrain[j,0], xtrain[j,1], l)
            ## if dk0 = k0.dK_dX(xtrain, xtrain, 1) :
            #dK[i,j] = func.dkdy(xtrain[i,0], xtrain[i,1], xtrain[j,0], xtrain[j,1], l)
    
    dK2 = np.zeros((len(x0train),len(x0train)))
    for i in range(dK2.shape[0]):
        for j in range(dK2.shape[1]):
            # if dk0_2 = k0.dK2_dXdX2(xtrain, xtrain, 0, 0) :
            dK2[i,j] = func.d2kdxdx0(xtrain[i,0], xtrain[i,1], xtrain[j,0], xtrain[j,1], l)
            ## if dk0_2 = k0.dK2_dXdX2(xtrain, xtrain, 0, 1) :
            #dK2[i,j] = func.d2kdxdy0(xtrain[i,0], xtrain[i,1], xtrain[j,0], xtrain[j,1], l)
            ## if dk0_2 = k0.dK2_dXdX2(xtrain, xtrain, 1, 0) :
            #dK2[i,j] = func.d2kdxdy0(xtrain[i,0], xtrain[i,1], xtrain[j,0], xtrain[j,1], l)
            ## if dk0_2 = k0.dK2_dXdX2(xtrain, xtrain, 1, 1) :
            #dK2[i,j] = func.d2kdydy0(xtrain[i,0], xtrain[i,1], xtrain[j,0], xtrain[j,1], l)
    
    # 2D plot :
    # plot the matrices
    plt.figure()
    plt.imshow(K)
    plt.title('K sympy' '\n' '2D')
    plt.figure()
    plt.imshow(k0_K)
    plt.title('K GPy' '\n' '2D')
    # plot one of the courbes
    plt.figure()
    plt.plot(K[49,:])
    plt.title('K sympy' '\n' '2D')
    plt.figure()
    plt.plot(k0_K[49,:])
    plt.title('K GPy' '\n' '2D')
    
    plt.figure()
    plt.imshow(dK)
    plt.title('1st derivative sympy' '\n' '2D')
    plt.figure()
    plt.imshow(dk0)
    plt.title('1st derivative GPy' '\n' '2D')
    plt.figure()
    plt.plot(dK[49,:])
    plt.title('1st derivative sympy' '\n' '2D')
    plt.figure()
    plt.plot(dk0[49,:])
    plt.title('1st derivative GPy' '\n' '2D')
    
    plt.figure()
    plt.imshow(dK2)
    plt.title('2nd derivative sympy' '\n' '2D')
    plt.figure()
    plt.imshow(dk0_2)
    plt.title('2nd derivative GPy' '\n' '2D')
    plt.figure()
    plt.plot(dK2[49,:])
    plt.title('2nd derivative sympy' '\n' '2D')
    plt.figure()
    plt.plot(dk0_2[49,:])
    plt.title('2nd derivative GPy' '\n' '2D')

    # Comparision (any dimension) :
    print(np.isclose(k0_K, K, rtol=1e-6),'\n')
    print(np.isclose(dk0, dK, rtol=1e-6),'\n')
    print(np.isclose(dk0_2, dK2, rtol=1e-6))
    
    

test_kern_Cosine_prod()



""" Mockup of a simulation with an independent parameter t in the output """

import numpy as np

x = np.loadtxt('parameters.in')
t = np.linspace(0, 2*np.pi, 100)

out = np.sin(5*x[0] + x[1]*t)

np.savetxt('result.out', out.reshape(1, -1))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Feb  4 08:34:52 2019

@author: Christopher Albert
"""
from profit.run import Worker


def rosenbrock(x, y, a, b):
    return (a - x)**2 + b * (y - x**2)**2


def func(r, u, v, a, b):
    return rosenbrock((r - 0.5) + u - 5, 1 + 3 * (v - 0.6), a, b)


@Worker.register('mockup')
class Mockup(Worker):
    def main(self):
        inputs = self.interface.input
        names = self.interface.input.dtype.names
        r = inputs['r'] if 'r' in names else 0.25
        u = inputs['u'] if 'u' in names else 5
        v = inputs['v'] if 'v' in names else 0.5
        a = inputs['a'] if 'a' in names else 1
        b = inputs['b'] if 'b' in names else 2
        self.interface.output['f'] = func(r, u, v, a, b)
        self.interface.done()



import numpy as np
import torch
import gpytorch


class ExactGPModel(gpytorch.models.ExactGP):
    def __init__(self, train_x, train_y, likelihood):
        super(ExactGPModel, self).__init__(train_x, train_y, likelihood)
        self.mean_module = gpytorch.means.LinearMean(train_x.shape[-1])
        #self.mean_module = gpytorch.means.ConstantMean()
        self.covar_module = gpytorch.kernels.ScaleKernel(gpytorch.kernels.RBFKernel())

    def forward(self, x):
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return gpytorch.distributions.MultivariateNormal(mean_x, covar_x)

# see https://docs.gpytorch.ai/en/v1.2.0/examples/01_Exact_GPs/Simple_GP_Regression.html
class GPyTorchSurrogate():
    def __init__(self):
        self.likelihood = gpytorch.likelihoods.GaussianLikelihood()

    def train(self, x, y, sigma_n=None, sigma_f=1e-6):
        self.ymean = np.mean(y)
        self.yscale = np.std(y)
        self.xtrain = torch.from_numpy(x).to(torch.float32)
        self.ytrain = torch.from_numpy((y - self.ymean)/self.yscale).to(torch.float32)
        self.ndim = x.shape[-1]

        self.m = ExactGPModel(self.xtrain, self.ytrain, self.likelihood)
        self.m.train()
        self.m.likelihood.train()


        # "Loss" for GPs - the marginal log likelihood
        mll = gpytorch.mlls.ExactMarginalLogLikelihood(self.m.likelihood, self.m)


        optimizer = torch.optim.Adam(self.m.parameters(), lr=0.1)  # Includes GaussianLikelihood parameters
        #optimizer = torch.optim.LBFGS(self.m.parameters())

        training_iter = 1000
        for i in range(training_iter):
            def closure():  # see http://sagecal.sourceforge.net/pytorch/index.html
                if torch.is_grad_enabled():
                    optimizer.zero_grad()
                output = self.m(self.xtrain)
                loss = -mll(output, self.ytrain)
                if loss.requires_grad:
                    loss.backward()
                if(i%100 == 0):
                    print('Iter %d/%d - Loss: %.3e   xscale: %.3e   yscale: %.3e   noise: %.3e' % (
                        i + 1, training_iter, loss.item(),
                        self.m.covar_module.base_kernel.lengthscale.item(),
                        self.m.covar_module.outputscale.item(),
                        self.m.likelihood.noise.item()
                    ))
                return loss
            optimizer.step(closure)

        self.trained = True

    def predict(self, x):
        self.m.eval()
        self.likelihood.eval()
        with torch.no_grad(), gpytorch.settings.fast_pred_var():
            pred = self.likelihood(self.m(torch.from_numpy(x).to(torch.float32)))
        return pred.mean.numpy()*self.yscale + self.ymean



import numpy as np
import time
from scipy.optimize import minimize
from profit.sur.backend.gpfunc import gpfunc
from profit.sur.backend.gp_functions import negative_log_likelihood_cholesky, solve_cholesky
from profit.sur.backend.python_kernels import RBF
from profit.util.halton import halton


def test_kern_sqexp():
    '''Fortran sqexp kernel gives correct result'''
    xdiff2 = np.linspace(0, 2, 128)
    tic = time.time()
    k1 = gpfunc.kern_sqexp(xdiff2)
    print(tic - time.time())
    tic = time.time()
    k2 = np.exp(-0.5*xdiff2)
    print(tic - time.time())

    assert np.array_equal(k1, k2)


def test_build_K():
    '''Matrix K is built correctly in Fortran build_K'''
    na = 3
    nb = 2
    xa = np.array([[0.0, 0.0], [0.0, 0.5], [-0.5, 0.1]])
    xb = np.array([[0.0, 0.0], [1.4, 1.0]])
    l = np.array([4.0, 5.0])

    K1 = np.empty((na, nb), order='F')
    gpfunc.build_k_sqexp(xa.T, xb.T, 1.0/l**2, K1)

    xdiff2 = np.empty((na, nb))
    for ka, xai in enumerate(xa):
        for kb, xbi in enumerate(xb):
            xdiff2[ka, kb] = np.sum(((xai - xbi)/l)**2)

    K2 = np.exp(-0.5*xdiff2)
    K3 = RBF(xa, xb, l)

    assert np.array_equal(K1, K2)
    assert np.array_equal(K1, K3)


def test_build_dKdth():
    '''Matrix derivatives dKdth is built correctly in Fortran'''
    na = 3
    nb = 2
    xa = np.array([[0.0, 0.0], [0.0, 0.5], [-0.5, 0.1]])
    xb = np.array([[0.0, 0.0], [1.4, 1.0]])
    l = np.array([4.0, 5.0])
    th = 1.0/l**2

    K = np.empty((na, nb), order='F')
    dK = np.empty((na, nb), order='F')
    gpfunc.build_k_sqexp(xa.T, xb.T, th, K)
    gpfunc.build_dkdth_sqexp(1, xa.T, xb.T, K, dK)

    dth = np.array([1e-6, 0.0])
    Kp = np.empty((na, nb), order='F')
    gpfunc.build_k_sqexp(xa.T, xb.T, th + 0.5*dth, Kp)
    Km = np.empty((na, nb), order='F')
    gpfunc.build_k_sqexp(xa.T, xb.T, th - 0.5*dth, Km)
    dKnum = (Kp - Km)/dth[0]
    assert(np.allclose(dK, dKnum, rtol=1e-8, atol=1e-9))

    gpfunc.build_dkdth_sqexp(2, xa.T, xb.T, K, dK)
    dth = np.array([0.0, 1e-6])
    Kp = np.empty((na, nb), order='F')
    gpfunc.build_k_sqexp(xa.T, xb.T, th + 0.5*dth, Kp)
    Km = np.empty((na, nb), order='F')
    gpfunc.build_k_sqexp(xa.T, xb.T, th - 0.5*dth, Km)
    dKnum = (Kp - Km)/dth[1]
    assert(np.allclose(dK, dKnum, rtol=1e-8, atol=1e-9))


def test_grad_nll():
    '''Gradient of nll matches finite difference approximation'''
    na = 3
    nb = 2
    xa = np.array([[0.0, 0.0], [0.0, 0.5], [-0.5, 0.1]])
    ya = np.sin(xa[:,0]) + np.cos(xa[:,1])
    xb = np.array([[0.0, 0.0], [1.4, 1.0]])
    l = np.array([3.0, 5.0])
    sigma_f = 1.0
    sigma_n = np.sqrt(0.1)
    hyp = np.hstack([l, [sigma_f, sigma_n]])

    K = RBF
    flik, dflik = negative_log_likelihood_cholesky(hyp, xa, ya, K, eval_gradient=True)

    # Testing length scales
    dl = 1e-6
    lm0 = l + [-0.5*dl, 0]
    flikm0 = negative_log_likelihood_cholesky(np.hstack([lm0, [sigma_f, sigma_n]]), xa, ya, K)
    lp0 = l + [0.5*dl, 0]
    flikp0 = negative_log_likelihood_cholesky(np.hstack([lp0, [sigma_f, sigma_n]]), xa, ya, K)
    l0m = l + [0, -0.5*dl]
    flik0m = negative_log_likelihood_cholesky(np.hstack([l0m, [sigma_f, sigma_n]]), xa, ya, K)
    l0p = l + [0, 0.5*dl]
    flik0p = negative_log_likelihood_cholesky(np.hstack([l0p, [sigma_f, sigma_n]]), xa, ya, K)
    dflik_num = np.array([flikp0-flikm0, flik0p-flik0m])/dl
    assert(np.allclose(dflik[:-2], dflik_num, rtol=1e-8, atol=1e-9))

    # Testing sig2f and sig2n
    ds2 = 1e-6
    flikm0 = negative_log_likelihood_cholesky(np.hstack([l, [sigma_f-0.5*ds2, sigma_n]]), xa, ya, K)
    flikp0 = negative_log_likelihood_cholesky(np.hstack([l, [sigma_f+0.5*ds2, sigma_n]]), xa, ya, K)
    flik0m = negative_log_likelihood_cholesky(np.hstack([l, [sigma_f, sigma_n-0.5*ds2]]), xa, ya, K)
    flik0p = negative_log_likelihood_cholesky(np.hstack([l, [sigma_f, sigma_n+0.5*ds2]]), xa, ya, K)

    dflik_num = np.array([flikp0-flikm0, flik0p-flik0m])/ds2
    print(dflik_num[-2])
    print(dflik[-2])
    assert(np.allclose(dflik[-2:], dflik_num, rtol=1e-8, atol=1e-9))


def test_fit_manual():
    ntrain = 128
    xtrain = halton(ntrain, 2)
    l = np.array([0.5, 0.5])
    sigf = 1.0
    sign = 1e-4
    hyp = np.hstack([l, sigf, sign])

    ytrain = np.sin(2*xtrain[:,0]) + np.cos(3*xtrain[:, 1])
    Ky = RBF(xtrain, xtrain, hyp[:-2], *hyp[-2:])
    L = np.linalg.cholesky(Ky)
    alpha = solve_cholesky(L, ytrain)

    n1test = 20
    n2test = 20
    ntest = n1test*n2test
    Xtest = np.mgrid[0:1:1j*n1test, 0:1:1j*n2test]
    xtest = Xtest.reshape([2,ntest]).T

    KstarT = RBF(xtest, xtrain, l, 1, 0)
    ymean = KstarT @ alpha
    yref = np.sin(2*xtest[:,0]) + np.cos(3*xtest[:,1])
    assert(np.allclose(ymean, yref, rtol=1e-3, atol=1e-3))


def test_optim():
    # TODO
    ntrain = 128
    xtrain = halton(ntrain, 2)
    l = np.array([0.5, 0.5])
    sigf = 1.0
    sign = 1e-4
    hyp = np.hstack([l, sigf, sign])

    ytrain = np.sin(2*xtrain[:, 0]) + np.cos(3*xtrain[:, 1])

    loghyp0 = np.log10(hyp)

    def cost(loghyp):
        h = 10**loghyp
        val, jac = negative_log_likelihood_cholesky(h, xtrain, ytrain, RBF, eval_gradient=True, log_scale_hyp=True)
        return val, h*np.log(10.0)*jac

    opt_res = minimize(cost, loghyp0, method='L-BFGS-B', jac=True, bounds=((-2, 2), (-2, 2), (-2, 2), (-4, 0)))


# TODO: test nll_chol

# TODO: port to gpfunc
# def test_kern_wendland4():
#     points_in = np.array([[0.0, 0.0], [0.0, 0.5], [-0.5, 0.1]])
#     points_out = np.array([[1.0, 1.4], [1.4, 1.0]])  # points where kern=0
#     h0 = 1.2
#     for x in points_in:
#         for x0 in points_in:
#             k1 = kernels.kern_wendland4(x, x0, h0)
#             k1a = kernels.kern_wendland4_multiscale(x, x0, np.array([h0, h0]))
#             r = np.sqrt(np.sum(((x-x0)/h0)**2))
#             k2 = (1.0 - r**4)*(1.0 + 4.0*r)
#             assert np.array_equal(k1, k2)
#             assert np.array_equal(k1a, k2)
#     for x in points_out:
#         for x0 in points_in:
#             k1 = kernels.kern_wendland4(x, x0, h0)
#             k1a = kernels.kern_wendland4_multiscale(x, x0, np.array([h0, h0]))
#             assert np.array_equal(k1, 0.0)
#             assert np.array_equal(k1a, 0.0)


# @pytest.mark.xfail(reason="Upstream bug https://github.com/pyccel/pyccel/issues/245")
# def test_kern_wendland4_product():
#     points_in = np.array([[0.0, 0.0], [0.0, 0.5], [-0.5, 0.1]])
#     points_out = np.array([[1.0, 2.0], [1.6, 1.0]])  # points where kern=0
#     h0 = 1.2
#     for x in points_in:
#         for x0 in points_in:
#             k1 = kernels.kern_wendland4_product(x, x0, h0)
#             k1a = kernels.kern_wendland4_product_multiscale(
#                 x, x0, np.array([h0, h0]))
#             dx = np.abs(x - x0)/h0
#             k2 = np.prod((1.0 - dx**4)*(1.0 + 4.0*dx))
#             assert np.array_equal(k1, k2)
#             assert np.array_equal(k1a, k2)
#     for x in points_out:
#         for x0 in points_in:
#             print(x, x0)
#             print(np.abs(x - x0)/h0)
#             k1 = kernels.kern_wendland4_product(x, x0, h0)
#             k1a = kernels.kern_wendland4_product_multiscale(
#                 x, x0, np.array([h0, h0]))
#             assert np.array_equal(k1, 0.0)
#             assert np.array_equal(k1a, 0.0)



import GPy
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns

sns.set()
import time
from IPython.display import display

# Definition of the dimension :
d = 1
# Number of samples ( training set ) :
n = 50
# Training input ( x ), should be defined as a matrix :
x = np.linspace(0, 1, n).reshape((n, 1))


# Definition of the function
def f(x):
    f = np.sin(2 * np.pi * x) + np.sin(4 * np.pi * x)
    return (f)


# Plot of the function f
plt.figure()
plt.plot(x, f(x))
plt.title('Plot of f(x) = $\sin(2 \pi x) + \sin(4 \pi x)$')
plt.xlabel('x')
plt.ylabel('f(x)')

# Definition of the training data (f_x = observations f(x))
f_x = f(x)

# the noise sigma_n
sigma_n = 1e-2
# Definition of the errors -> we assume that the noise on observations follows an
# independent , identically distributed Gaussian distribution with zero mean and
# variance sigma_n ^2
epsilon = sigma_n * np.random.randn(n).reshape((-1, 1))
# Observed target variable (f(x) + epsilon)
y = f_x + epsilon
# Plot of the error distribution
plt.figure()
sns.distplot(epsilon)
plt.title('Distribution of $\epsilon$ and the interpolated density')
plt.xlabel('$\epsilon$')
plt.ylabel('Number of values')
# define number of test data points
n_star = 30
# define input test data x_star
x_star = np.linspace(0, 1, n_star).reshape((n_star, 1))
# set the hyper - parameters l and sigma_f
l = 0.2
sigma_f = 1

# Definition of the squared exponential kernel
t = time.time()
kernel = GPy.kern.RBF(input_dim=d, variance=sigma_f ** 2, lengthscale=l)
elapsed = time.time() - t
print(elapsed)
# Plot of the kernel function
kernel.plot()
# Build the model
t = time.time()
m = GPy.models.GPRegression(x, y, kernel, noise_var=0.1 ** 2)
elapsed = time.time() - t
print(elapsed)
# Display the model
display(m)

# plot the posterior GP
plt.figure()
t = time.time()
GPy.plotting.gpy_plot.gp_plots.plot(m)
elapsed = time.time() - t
print(elapsed)
# f , varf
# vary = varf + sign **2
# Compute the covariance matrices
t = time.time()
K = kernel.K(x, x)
K_star2 = kernel.K(x_star, x_star)
K_star = kernel.K(x_star, x)
elapsed = time.time() - t
print(elapsed)
# compute C
C = np.block([[K + sigma_n ** 2 * np.eye(n), np.transpose(K_star)], [K_star, K_star2]])
# plot function f(x)
plt.figure()
plt.plot(x, f_x, "r")
# print(x.shape)
plt.legend('f')
# plot of n_prior samples from prior distribution (100 samples)
n_prior = 100
for i in range(0, n_prior):
    f_star = np.random.multivariate_normal(np.zeros(n_star), K_star2)
    plt.plot(x_star, f_star, 'b', linewidth='0.5')
plt.title('Plot of f and 100 Samples from the Prior distribution')
plt.xlabel('x')
# Compute posterior mean and covariance.
t = time.time()
f_bar_star, cov_f_star = m.predict(x_star, full_cov=True)
elapsed = time.time() - t
print(elapsed)
plt.figure()
# Plot of n_prior samples from the posterior distribution and of the true function
for i in range(n_prior):
    f_posterior = np.random.multivariate_normal(f_bar_star[:, 0], cov_f_star)
    plt.plot(x_star, f_posterior)
plt.fill_between(x_star.flatten(),  # x
                 (f_bar_star.flatten() + 2 * np.sqrt(np.diag(cov_f_star))),  # y1
                 (f_bar_star.flatten() - 2 * np.sqrt(np.diag(cov_f_star))))  # y2
plt.title(' 100 Samples from Posterior Distribution and 95% confidence interval ')
plt.xlabel('x ')
plt.ylabel('y ')

# Since we don ’t want one of the variance to become negative during the optimization
# , we constrain all parameters to be positive before running the optimisation
m.constrain_positive('')  # '' is a regex matching all parameter names
t = time.time()
m.optimize(messages=True)
elapsed = time.time() - t
print(elapsed)
display(m)
# Plot of the optimized posterior GP
t = time.time()
plt.figure()
GPy.plotting.gpy_plot.gp_plots.plot(m)
elapsed = time.time() - t
print(elapsed)
plt.title(' Optimised GPR of f ')
plt.xlabel('x ')
plt.ylabel('f ( x ) ')
plt.show()



#%%
import numpy as np
import matplotlib.pyplot as plt
from scipy.sparse.linalg.eigen import eigsh
from sklearn.model_selection import train_test_split
from sklearn.decomposition import PCA, KernelPCA

#%%
eps = 0.00
def pend_data(z0, dt, nt):
    z = np.zeros([nt+1, 2])
    z[0, :] = z0
    for k in range(nt):
        qmid = z[k, 0] + 0.5*dt*z[k, 1]
        z[k+1, 1] = z[k, 1] - dt*np.sin(qmid)
        z[k+1, 0] = qmid + 0.5*dt*z[k+1, 1]

    return z

ntrain = 100
nt = 300
qtrain = np.empty([ntrain, nt+1])
ptrain = np.empty([ntrain, nt+1])
p0train = 1.5 + 0.01*np.random.randn(ntrain)
q0train = 1.2 + 0.01*np.random.randn(ntrain)
#p0train = .5*np.random.randn(ntrain)
#q0train = 1.5*np.random.randn(ntrain)
#p0train = 5.0*(np.random.rand(ntrain) - 0.5)
#q0train = 2.5*(np.random.rand(ntrain) - 0.5)
for ind, p0 in enumerate(p0train):
    z = pend_data([q0train[ind], p0], 0.1, nt)
    qtrain[ind, :] = z[:, 0]
    ptrain[ind, :] = z[:, 1]

plt.figure()
plt.plot(qtrain.T, ptrain.T)
# %%
neig = 2
qmean = np.mean(qtrain, 0)
dqtrain = qtrain - qmean
pmean = np.mean(ptrain, 0)
dptrain = ptrain - pmean
#pca = KernelPCA(n_components=neig, kernel='linear', fit_inverse_transform=True)
pca = PCA(n_components=neig)
atrain = pca.fit_transform(dqtrain)
#print(f'error: {np.sqrt(pca.noise_variance_):.1e}')
#%%
phi = np.empty([neig, nt+1])
for k in range(neig):  # Compute k-th eigenfunction
    x = np.zeros(neig)
    x[k] = 1.0
    phi[k, :] = pca.inverse_transform(x.reshape(1,-1))

plt.figure()
plt.plot(phi.T)
#plt.plot(qtrain[0,:])
plt.title('Eigenfunctions')

#%%
plt.figure()
plt.plot(atrain[:,0], atrain[:,1], 'x')
plt.axis('equal')
plt.title('Weights')

# %%
from scipy.sparse.linalg.eigen import eigsh

neig=5

M = np.empty([2*ntrain, 2*ntrain])

M[:ntrain, :ntrain] = dqtrain @ dqtrain.T
M[:ntrain, ntrain:] = dqtrain @ dptrain.T
M[ntrain:, :ntrain] = dptrain @ dqtrain.T
M[ntrain:, ntrain:] = dptrain @ dptrain.T
w, Q = eigsh(M, neig)
Q = Q[:,:-neig-1:-1]
w = w[:-neig-1:-1]

plt.figure()
plt.semilogy(w)

for i in range(len(w)):
    plt.figure()
    qeig = Q[:ntrain,i] @ dqtrain
    peig = Q[ntrain:,i] @ dptrain
    plt.figure()
    plt.plot(qeig, peig)
    plt.xlabel('qeig')
    plt.ylabel('peig')

# %%
U, S, V = np.linalg.svd(np.vstack([dqtrain, dptrain]))
# Full SVD: nxn @ diag(nxn) @ nxp
# Partial SVD: n x neig @ diag(neig x neig) @ neig x p

# %%



from sympy import *
from sympy.utilities.autowrap import autowrap, ufuncify
from sympy.utilities.lambdify import lambdastr, lambdify, implemented_function
import shutil

tmp = '.'
#try:
#    shutil.rmtree(tmp)
#except:
#    pass

#%% # Kernel functions (kernel, 1st and second derivatives)
x, y, xa, ya, xb, yb, l, lx, ly = symbols(r'x y x_a y_a x_b, y_b, l, lx, ly')


kern_sqexp = exp(-x**2/(2*l**2))
kern_sqexp_sin = exp(-sin(x/2)**2/(2*l**2))

kern_y = kern_sqexp.subs(x, ya-yb).subs(l, ly)
kern_x = kern_sqexp.subs(x, xa-xb).subs(l, lx)
kern = (kern_x*kern_y).simplify()

dkdxa = diff(kern, xa).simplify()
dkdxb = diff(kern, xb).simplify()
dkdya = diff(kern, ya).simplify()
dkdyb = diff(kern, yb).simplify()
dkdxadxb = diff(kern, xa, xb).simplify()
dkdyadyb = diff(kern, ya, yb).simplify()
dkdxadyb = diff(kern, xa, yb).simplify()

d3kdxdx0dy0 = diff(kern, xa, xb, yb).simplify()
d3kdydy0dy0 = diff(kern, ya, yb, yb).simplify()
d3kdxdy0dy0 = diff(kern, xa, yb, yb).simplify()

dkdlx = diff(kern, lx).simplify()
dkdly = diff(kern, ly).simplify()

d3kdxdx0dlx = diff(kern, xa, xb, lx).simplify()
d3kdydy0dlx = diff(kern, ya, yb, lx).simplify()
d3kdxdy0dlx = diff(kern, xa, yb, lx).simplify()

d3kdxdx0dly = diff(kern, xa, xb, ly).simplify()
d3kdydy0dly = diff(kern, ya, yb, ly).simplify()
d3kdxdy0dly = diff(kern, xa, yb, ly).simplify()

#%%
from sympy.utilities.codegen import codegen
seq = (xa, ya, xb, yb, lx, ly)
[(name, code), (h_name, header)] = \
codegen([('kern_num', kern),
         ('dkdx_num', dkdxa),
         ('dkdy_num', dkdya),
         ('dkdx0_num', dkdxb),
         ('dkdy0_num', dkdyb),
         ('d2kdxdx0_num', dkdxadxb),
         ('d2kdydy0_num', dkdyadyb),
         ('d2kdxdy0_num', dkdxadyb),
         ('d3kdxdx0dy0_num', d3kdxdx0dy0),
         ('d3kdydy0dy0_num', d3kdydy0dy0),
         ('d3kdxdy0dy0_num', d3kdxdy0dy0),
         ('dkdlx_num', dkdlx),
         ('dkdly_num', dkdly),
         ('d3kdxdx0dlx_num', d3kdxdx0dlx),
         ('d3kdydy0dlx_num', d3kdydy0dlx),
         ('d3kdxdy0dlx_num', d3kdxdy0dlx),
         ('d3kdxdx0dly_num', d3kdxdx0dly),
         ('d3kdydy0dly_num', d3kdydy0dly),
         ('d3kdxdy0dly_num', d3kdxdy0dly),
         ], "F95", "kernels",
        argument_sequence=seq,
        header=False, empty=False)
with open(name, 'w') as f:
    f.write(code)



import dash
import dash_core_components as dcc
import dash_html_components as html
import plotly.graph_objs as go
from dash.dependencies import Input, Output, State, MATCH, ALL
from math import log10
import numpy as np
from profit.util import load
from profit.sur import Surrogate
from matplotlib import cm as colormaps
from matplotlib.colors import to_hex as color2hex

def init_app(config):
    external_stylesheets = ['https://codepen.io/chriddyp/pen/bWLwgP.css']

    app = dash.Dash(__name__, external_stylesheets=external_stylesheets)
    server = app.server
    app.config.suppress_callback_exceptions = False

    indata = load(config['files']['input']).flatten()
    outdata = load(config['files']['output']).flatten()

    # data = pd.concat([indata, outdata], 1) # TODO: data in table

    invars = indata.dtype.names
    outvars = outdata.dtype.names
    dd_opts_in = [{'label': invar, 'value': invar} for invar in invars]
    dd_opts_out = [{'label': outvar, 'value': outvar} for outvar in outvars]

    col_width = 400
    txt_width = 100
    dd_width = 250
    log_width = 50
    graph_height = 620
    txt_check_width = 50
    check_txt_width = txt_width-txt_check_width
    ax_opt_tit_sty = {'width': col_width}
    ax_opt_txt_sty = {'width': txt_width}
    ax_opt_log_sty = {'width': log_width}
    dd_sty = {'width': dd_width}
    axis_options_div_style = {'display': 'flex', 'align-items': 'center', 'height':36, 'padding': 1}
    fit_opt_txt_sty = {'width': txt_width}
    headline_sty = {'text-align': 'center', 'display': 'block', 'width': col_width-25}
    input_div_sty = {'height': 40}
    input_sty = {'width': 125}

    # try to load model with 'save' and 'fit' config option
    path = config['fit']['save'] or config['fit']['load']
    try:
        sur = Surrogate.load_model(path)
    except (TypeError, FileNotFoundError):
            print('Model could not be loaded')

    app.layout = html.Div(children=[
        html.Table(children=[html.Tr(children=[
            html.Td(id='axis-options', style={'width': '20%'}, children=[
                html.Div(dcc.RadioItems(
                    id='graph-type',
                    options=[{'label': i, 'value': i} for i in ['1D', '2D', '2D contour', '3D']],
                    value='1D',
                    labelStyle={'display': 'inline-block'})),
                html.Div(id='header-opt', children=[html.B("Axis options:", style=headline_sty)], style=ax_opt_tit_sty),
                html.Div(id='invar-1-div', style=axis_options_div_style, children=[
                    html.B('x: ', style=ax_opt_txt_sty),
                    dcc.Dropdown(id='invar', options=dd_opts_in, value=invars[0], style=dd_sty),
                    dcc.Checklist(id='invar-1-log', options=[{'label': 'log', 'value': 'log'}], style=ax_opt_log_sty),
                ]),
                html.Div(id='invar-2-div', style=axis_options_div_style, children=[
                    html.B('y: ', style=ax_opt_txt_sty),
                    dcc.Dropdown(
                        id='invar_2',
                        options=dd_opts_in,
                        value=invars[1] if len(invars) > 1 else invars[0],
                        style=dd_sty, ),
                    dcc.Checklist(
                        id='invar-2-log',
                        options=[{'label': 'log', 'value': 'log'}],
                        style=ax_opt_log_sty, ),
                ]),
                html.Div(id='invar-3-div', style=axis_options_div_style, children=[
                    html.B('z: ', style=ax_opt_txt_sty),
                    dcc.Dropdown(
                        id='invar_3',
                        options=dd_opts_in,
                        value=invars[2] if len(invars) > 2 else invars[0],
                        style=dd_sty,
                    ),
                    dcc.Checklist(
                        id='invar-3-log',
                        options=[{'label': 'log', 'value': 'log'}],
                        style=ax_opt_log_sty,
                    ),
                ]),
                html.Div(id='outvar-div', style=axis_options_div_style, children=[
                    html.B('output: ', style=ax_opt_txt_sty),
                    dcc.Dropdown(
                        id='outvar',
                        options=dd_opts_out,
                        value=outvars[0],
                        style=dd_sty, ),
                    dcc.Checklist(
                        id='outvar-log',
                        options=[{'label': 'log', 'value': 'log'}],
                        style=ax_opt_log_sty, ),
                ]),
                html.Div(id='color-div', style=axis_options_div_style, children=[
                    html.B("color: ", style={'width': txt_check_width}),
                    dcc.Checklist(
                        id='color-use',
                        options=[{'label': '', 'value': 'true'}],
                        style={'width': check_txt_width},
                        value=['true'], ),
                    dcc.Dropdown(
                        id='color-dropdown',
                        options=[{'label': 'OUTPUT', 'value': 'OUTPUT'}] + dd_opts_in + dd_opts_out,
                        value='OUTPUT',
                        style=dd_sty, ),
                ]),
                html.Div(id='error-div', style=axis_options_div_style, children=[
                    html.B("error: ", style={'width': txt_check_width}),
                    dcc.Checklist(
                        id='error-use',
                        options=[{'label': '', 'value': 'true'}],
                        style={'width': check_txt_width}, ),
                    dcc.Dropdown(
                        id='error-dropdown',
                        options=dd_opts_out,
                        value=outvars[-1],
                        style=dd_sty, ),
                ]),
                html.Div(id='fit-opt', children=html.B("Fit options:", style=headline_sty), style=ax_opt_tit_sty),
                html.Div(id='fit-use-div', style=axis_options_div_style, children=[
                    html.B("display fit:", style=fit_opt_txt_sty),
                    dcc.Checklist(
                        id='fit-use',
                        options=[{'label': '', 'value': 'show'}],
                        labelStyle={'display': 'inline-block'}, ),
                ]),
                html.Div(id='fit-multiinput-div', style=axis_options_div_style, children=[
                    html.B("multi-fit:", style=fit_opt_txt_sty),
                    dcc.Dropdown(
                        id='fit-multiinput-dropdown',
                        options=dd_opts_in,
                        value=invars[-1],
                        style=dd_sty, ),
                ]),
                html.Div(id='fit-number-div', style=axis_options_div_style, children=[
                    html.B("#fits:", style=fit_opt_txt_sty),
                    dcc.Input(id='fit-number', type='number', value=1, min=1),
                ]),
                html.Div(id='fit-conf-div', style=axis_options_div_style, children=[
                    html.B("\u03c3-confidence:", style=fit_opt_txt_sty),
                    dcc.Input(id='fit-conf', type='number', value=2, min=0),
                ]),
                html.Div(id='fit-noise-div', style=axis_options_div_style, children=[
                    dcc.Checklist(
                        id='fit-var',
                        options=[{'label': 'add noise covariance', 'value': 'add'}],
                        style={'margin-left': txt_width},
                    )
                ]),
                html.Div(id='fit-color-div', style=axis_options_div_style, children=[
                    html.B("fit-color:", style=fit_opt_txt_sty),
                    dcc.RadioItems(
                        id='fit-color',
                        options=[{'label': 'output', 'value': 'output'},
                                 {'label': 'multi-fit', 'value': 'multi-fit'},
                                 {'label': 'marker-color', 'value': 'marker-color'}],
                        value='output',
                        labelStyle={'display': 'inline-block'}, ),
                ]),
                html.Div(id='fit-opacity-div', style=axis_options_div_style, children=[
                    html.B("fit-opacity:", style=fit_opt_txt_sty),
                    html.Div(style={'width': col_width-txt_width}, children=[
                        dcc.Slider(
                            id='fit-opacity',
                            min=0,
                            max=1,
                            step=0.1,
                            value=0.5,
                            marks={i: {'label': f'{100 * i:.0f}%'} for i in [0, 0.2, 0.4, 0.6, 0.8, 1]},
                        ),
                    ]),
                ]),
                html.Div(id='fit-sampling-div', style=axis_options_div_style, children=[
                    html.B("#points:", style=fit_opt_txt_sty),
                    dcc.Input(id='fit-sampling', type='number', value=50, min=1, debounce=True, style={'appearance': 'textfield'}),
                ]),
            ]),
            html.Td(id='graph', style={'width': '80%'}, children=[html.Div(dcc.Graph(id='graph1'))]),
        ])]),
        html.Div(html.Table(id='filters', children=[html.Tr([
            html.Td(html.Div([
                dcc.Dropdown(
                    id='filter-dropdown',
                    options=dd_opts_in,
                    value=invars[0],
                    style={'width': 200, 'margin-right': 10},
                ),
                html.Button("Add Filter", id='add-filter', n_clicks=0),
            ], style={'display': 'flex'})),
            html.Td(html.Button("Clear Filter", id='clear-filter', n_clicks=0)),
            html.Td(html.Button("Clear all Filter", id='clear-all-filter', n_clicks=0)),
            html.Td(dcc.Slider(id='scale-slider',
                               min=-0.5, max=0.5,
                               value=0, step=0.01,
                               marks={i: f'{100*i:.0f}%' for i in [-0.5, -0.25, 0, 0.25, 0.5]},
                               ),
                    style={'width': 500}
            ),
            html.Td(html.Button("Scale Filter span", id='scale', n_clicks=0)),
        ])])),
        html.Div(html.Table(id='param-table', children=[
            html.Thead(id='param-table-head', children=[
                html.Tr(children=[
                    html.Th("Parameter", style=input_sty),
                    html.Th("log"),
                    html.Th("Slider", style={'width': 300}),
                    html.Th("Range (min/max)"),
                    html.Th("center/span"),
                    html.Th("filter active"),
                    html.Th("#digits"),
                    html.Th("reset"),
                ]),
            ]),
            html.Tbody(id='param-table-body', children=[
                html.Tr(children=[
                    html.Td(html.Div(id='param-text-div', children=[])),
                    html.Td(html.Div(id='param-log-div', children=[])),
                    html.Td(html.Div(id='param-slider-div', children=[])),
                    html.Td(html.Div(id='param-range-div', children=[])),
                    html.Td(html.Div(id='param-center-div', children=[])),
                    html.Td(html.Div(id='param-active-div', children=[])),
                    html.Td(html.Div(id='param-digits-div', children=[])),
                    html.Td(html.Div(id='param-reset-div', children=[])),
                ]),
            ]),
        ])),
    ])


    @app.callback(
        [Output('param-text-div', 'children'),
         Output('param-log-div', 'children'),
         Output('param-slider-div', 'children'),
         Output('param-range-div', 'children'),
         Output('param-center-div', 'children'),
         Output('param-active-div', 'children'),
         Output('param-digits-div', 'children'),
         Output('param-reset-div', 'children'), ],
        [Input('add-filter', 'n_clicks'),
         Input('clear-filter', 'n_clicks'),
         Input('clear-all-filter', 'n_clicks')],
        [State('filter-dropdown', 'value'),
         State('param-text-div', 'children'),
         State('param-log-div', 'children'),
         State('param-slider-div', 'children'),
         State('param-range-div', 'children'),
         State('param-center-div', 'children'),
         State('param-active-div', 'children'),
         State('param-digits-div', 'children'),
         State('param-reset-div', 'children')],
    )
    def add_filterrow(n_clicks, clear, clear_all, filter_dd, text, log, slider, range_div, center_div, active_div, dig_div, reset_div):
        ctx = dash.callback_context
        trigger_id = ctx.triggered[0]["prop_id"].split(".")[0]
        if trigger_id == 'clear-all-filter':
            return [], [], [], [], [], [], [], []
        elif trigger_id == 'clear-filter':
            for i in range(len(text)):
                if text[i]['props']['children'][0] == filter_dd:
                    text.pop(i)
                    log.pop(i)
                    slider.pop(i)
                    range_div.pop(i)
                    center_div.pop(i)
                    active_div.pop(i)
                    dig_div.pop(i)
                    reset_div.pop(i)
                    break
        elif trigger_id == 'add-filter':
            for i in range(len(text)):
                if text[i]['props']['children'][0] == filter_dd:
                    return text, log, slider, range_div, center_div, active_div, dig_div, reset_div
            ind = invars.index(filter_dd)
            txt = filter_dd
            new_text = html.Div(id={'type': 'dyn-text', 'index': ind}, children=[txt], style={**input_div_sty, **input_sty})
            new_log = html.Div(id={'type': 'dyn-log', 'index': ind}, style=input_div_sty, children=[
                dcc.Checklist(id={'type': 'param-log', 'index': ind}, options=[{'label': '', 'value': 'log'}])], )
            new_slider = html.Div(id={'type': 'dyn-slider', 'index': ind}, style=input_div_sty, children=[
                create_slider(txt)], )
            new_range = html.Div(id={'type': 'dyn-range', 'index': ind}, style=input_div_sty, children=[
                dcc.Input(id={'type': 'param-range-min', 'index': ind}, type='number', debounce=True, style={**input_sty, 'appearance': 'textfield'}),
                dcc.Input(id={'type': 'param-range-max', 'index': ind}, type='number', debounce=True, style={**input_sty, 'appearance': 'textfield'}),
            ], )
            new_center = html.Div(id={'type': 'dyn-center', 'index': ind}, style=input_div_sty, children=[
                dcc.Input(id={'type': 'param-center', 'index': ind}, type='number', debounce=True, style={**input_sty, 'appearance': 'textfield'}),
                dcc.Input(id={'type': 'param-span', 'index': ind}, type='number', debounce=True, style={**input_sty, 'appearance': 'textfield'}),
            ], )
            new_active = html.Div(id={'type': 'dyn-active', 'index': ind},
                                  style={**input_div_sty, 'text-align': 'center'},
                                  children=[
                                      dcc.Checklist(id={'type': 'param-active', 'index': ind},
                                                    options=[{'label': '', 'value': 'act'}],
                                                    value=['act'], )
                                  ])
            new_dig = html.Div(id={'type': 'dyn-dig', 'index': ind}, children=[
                dcc.Input(id={'type': 'param-dig', 'index': ind}, type='number', value=5, min=0, style={'width':100})
            ])
            new_reset = html.Div(id={'type': 'dyn-reset', 'index': ind}, children=[
                html.Button("reset", id={'type': 'param-reset', 'index': ind}, n_clicks=0)
            ])
            text.append(new_text)
            log.append(new_log)
            slider.append(new_slider)
            range_div.append(new_range)
            center_div.append(new_center)
            active_div.append(new_active)
            dig_div.append(new_dig)
            reset_div.append(new_reset)
        return text, log, slider, range_div, center_div, active_div, dig_div, reset_div


    @app.callback(
        [Output({'type': 'param-range-min', 'index': MATCH}, 'step'),
         Output({'type': 'param-range-max', 'index': MATCH}, 'step'),
         Output({'type': 'param-center', 'index': MATCH}, 'step'),
         Output({'type': 'param-span', 'index': MATCH}, 'step'),
         Output({'type': 'param-slider', 'index': MATCH}, 'step'), ],
        Input({'type': 'param-dig', 'index': MATCH}, 'value')
    )
    def update_step(dig):
        """ Function to update an synchronise step-sizes throughout the filter-table.

        Args:
            dig (int): Number of digits to be used. Selected by the user via a 'dcc.Input'-layout-element.

        Returns:
            step: Step-size for the 4 'dcc.Input'-Elements and the slider step.

        """
        step = 10**(-dig)
        return step, step, step, step, step


    @app.callback(
        [Output({'type': 'param-range-min', 'index': MATCH}, 'value'),
         Output({'type': 'param-range-max', 'index': MATCH}, 'value'),
         Output({'type': 'param-slider', 'index': MATCH}, 'value'),
         Output({'type': 'param-slider', 'index': MATCH}, 'min'),
         Output({'type': 'param-slider', 'index': MATCH}, 'max'),
         Output({'type': 'param-center', 'index': MATCH}, 'value'),
         Output({'type': 'param-span', 'index': MATCH}, 'value'),
         Output({'type': 'param-slider', 'index': MATCH}, 'marks'), ],
        [Input('param-text-div', 'children'),
         Input({'type': 'param-log', 'index': MATCH}, 'value'),
         Input({'type': 'param-range-min', 'index': MATCH}, 'value'),
         Input({'type': 'param-range-max', 'index': MATCH}, 'value'),
         Input({'type': 'param-slider', 'index': MATCH}, 'value'),
         Input({'type': 'param-center', 'index': MATCH}, 'value'),
         Input({'type': 'param-span', 'index': MATCH}, 'value'),
         Input('scale', 'n_clicks'),
         Input({'type': 'param-dig', 'index': MATCH}, 'value'),
         Input({'type': 'param-reset', 'index': MATCH}, 'n_clicks'), ],
        [State({'type': 'param-slider', 'index': MATCH}, 'id'),
         State('scale-slider', 'value'),
         State({'type': 'param-slider', 'index': MATCH}, 'marks'), ]
    )
    def update_dyn_slider_range(text_div, log_act, dyn_min, dyn_max, slider_val, center, span, scale, dig, reset, id, scale_slider, marks):
        ctx = dash.callback_context
        try:
            trigger_id = ctx.triggered[0]["prop_id"].split('}')[0].split(',')[1].split(':')[1]
        except IndexError:
            trigger_id = ctx.triggered[0]["prop_id"]

        mark_lim = [float(i) for i in list(marks.keys())]

        data_in = indata[invars[id['index']]]
        data_min_0 = min(data_in[data_in > 0])

        if trigger_id == '"param-log"' and log_act != ['log']:
            dyn_min = 10**dyn_min
            dyn_max = 10**dyn_max
            slider_val = [10**val for val in slider_val]
            mark_lim = [10 ** lim for lim in mark_lim]
            if min(data_in) < 0 and slider_val[0] > 0:
                mark_lim[0] = min(data_in)
            span = (slider_val[1] - slider_val[0])/2
            center = (slider_val[0] + slider_val[1])/2

        if trigger_id != '"param-log"' and log_act == ['log']:
            dyn_min = 10 ** dyn_min
            dyn_max = 10 ** dyn_max
            slider_val = [10 ** val for val in slider_val]
            mark_lim = [10 ** lim for lim in mark_lim]
            if min(data_in) < 0 and slider_val[0] > 0:
                mark_lim[0] = min(data_in)

        if trigger_id == '"param-reset"':
            slider_val = [min(data_in), max(data_in)]

        if ctx.triggered[0]["prop_id"] == "scale.n_clicks":
            # print('scale')
            span = span * (1 + scale_slider)
            if log_act == ['log']:
                trigger_id = '"param-span"'
            else:
                dyn_min = center - span
                dyn_max = center + span
                slider_val = [dyn_min, dyn_max]

        if trigger_id == '"param-center"' or trigger_id == '"param-span"' and (center and span):
            # print('center')
            if log_act == ['log']:
                dyn_min = 10**(center-span)
                dyn_max = 10**(center+span)
            else:
                dyn_min = center - span
                dyn_max = center + span
            slider_val = [dyn_min, dyn_max]
        elif (trigger_id == '"param-range-min"' or trigger_id == '"param-range-max"') and (
                dyn_min is not None and dyn_max is not None):
            # print('range')
            slider_val = [dyn_min, dyn_max]
            span = (slider_val[1] - slider_val[0]) / 2
            center = (slider_val[0] + slider_val[1]) / 2
        elif slider_val:
            # print('slider')
            dyn_min = slider_val[0]
            dyn_max = slider_val[1]
            span = (slider_val[1] - slider_val[0]) / 2
            center = (slider_val[0] + slider_val[1]) / 2

        if log_act == ['log']:
            # log values
            try:
                log_dyn_min = log10(dyn_min)
            except ValueError:
                log_dyn_min = log10(data_min_0)
            log_dyn_max = log10(dyn_max)
            log_slider_val = [log_dyn_min, log_dyn_max]
            try:
                log_mark_lim = [log10(mark) for mark in mark_lim]
            except ValueError:
                log_mark_lim = [log10(data_min_0), log10(mark_lim[1])]
            log_span = (log_slider_val[1] - log_slider_val[0])/2
            log_center = log_slider_val[0] + log_span
            log_marks = {log_mark_lim[0]: str(round(log_mark_lim[0], dig)),
                         log_mark_lim[1]: str(round(log_mark_lim[1], dig))}
            return round(log_dyn_min, dig), round(log_dyn_max, dig), log_slider_val, log_mark_lim[0], log_mark_lim[1],\
                   round(log_center, dig), round(log_span, dig), log_marks
        else:
            marks = {mark_lim[0]: str(round(mark_lim[0], dig)), mark_lim[1]: str(round(mark_lim[1], dig))}
            return round(dyn_min, dig), round(dyn_max, dig), slider_val, mark_lim[0], mark_lim[1], \
                   round(center, dig), round(span, dig), marks


    @app.callback(
        [Output('invar-2-div', 'style'),
         Output('invar-3-div', 'style'),
         Output('color-div', 'style'),
         Output('error-div', 'style'),
         Output('fit-use-div', 'style'),
         Output('fit-multiinput-div', 'style'),
         Output('fit-number-div', 'style'),
         Output('fit-number', 'value'),
         Output('fit-conf-div', 'style'),
         Output('fit-noise-div', 'style'),
         Output('fit-color-div', 'style'),
         Output('fit-opacity-div', 'style'), ],
        [Input('graph-type', 'value'), ],
        [State('fit-number', 'value'), ]
    )
    def div_visibility(graph_type, fits):
        hide = axis_options_div_style.copy()
        hide['visibility'] = 'hidden'
        show = axis_options_div_style.copy()
        show['visibility'] = 'visible'
        if graph_type == '1D':
            return hide, hide, show, show, show, show, show, fits, show, show, hide, show
        if graph_type == '2D':
            if len(invars) <= 2:
                return show, hide, show, show, show, hide, hide, 1, show, show, show, show
            else:
                return show, hide, show, show, show, show, show, fits, show, show, show, show
        if graph_type == '2D contour':
            return show, hide, show, hide, hide, hide, hide, fits, hide, hide, hide, hide
        if graph_type == '3D':
            return show, show, hide, hide, show, hide, show, fits, hide, hide, hide, show
        else:
            return show, show, show, show, show, show, show, fits, show, show, show, show


    @app.callback(
        Output('graph1', 'figure'),
        [Input('invar', 'value'),
         Input('invar_2', 'value'),
         Input('invar_3', 'value'),
         Input('outvar', 'value'),
         Input('invar-1-log', 'value'),
         Input('invar-2-log', 'value'),
         Input('invar-3-log', 'value'),
         Input('outvar-log', 'value'),
         Input({'type': 'param-slider', 'index': ALL}, 'value'),
         Input('graph-type', 'value'),
         Input('color-use', 'value'),
         Input('color-dropdown', 'value'),
         Input('error-use', 'value'),
         Input('error-dropdown', 'value'),
         Input({'type': 'param-active', 'index': ALL}, 'value'),
         Input('fit-use', 'value'),
         Input('fit-multiinput-dropdown', 'value'),
         Input('fit-number', 'value'),
         Input('fit-conf', 'value'),
         Input('fit-var', 'value'),
         Input('fit-color', 'value'),
         Input('fit-opacity', 'value'),
         Input('fit-sampling', 'value'), ],
        [State({'type': 'param-slider', 'index': ALL}, 'id'),
         State({'type': 'param-center', 'index': ALL}, 'value'),
         State({'type': 'param-log', 'index': ALL}, 'value')],
    )
    def update_figure(invar, invar_2, invar_3, outvar, invar1_log, invar2_log, invar3_log, outvar_log, param_slider,
                      graph_type, color_use, color_dd, error_use, error_dd, filter_active, fit_use, fit_dd, fit_num, fit_conf, add_noise_var, fit_color,
                      fit_opacity, fit_sampling, id_type, param_center, param_log):
        for i in range(len(param_slider)):
            if param_log[i] == ['log']:
                param_slider[i] = [10**val for val in param_slider[i]]
                param_center[i] = 10**param_center[i]
        if invar is None:
            return go.Figure()
        sel_y = np.full((len(outdata),), True)
        dds_value = []
        for iteration, values in enumerate(param_slider):
            dds_value.append(invars[id_type[iteration]['index']])
            # filter for minimum
            sel_y_min = np.array(indata[dds_value[iteration]] >= param_slider[iteration][0])
            # filter for maximum
            sel_y_max = np.array(indata[dds_value[iteration]] <= param_slider[iteration][1])
            # print('iter ', iteration, 'filer', filter_active[iteration][0])
            if filter_active != [[]]:
                if filter_active[iteration] == ['act']:
                    sel_y = sel_y_min & sel_y_max & sel_y
        if graph_type == '1D':
            fig = go.Figure(
                data=[go.Scatter(
                    x=indata[invar][sel_y],
                    y=outdata[outvar][sel_y],
                    mode='markers',
                    name='data',
                    error_y=dict(type='data', array=outdata[error_dd][sel_y], visible= error_use == ['true']),
                    # text=[(invar, outvar) for i in range(len(indata[invar][sel_y]))],
                    # hovertemplate=" %{text} <br> %{x} <br> %{y}",
                )],
                layout=go.Layout(xaxis=dict(title=invar, rangeslider=dict(visible=True)), yaxis=dict(title=outvar))
            )
            if fit_use == ['show']:
                mesh_in, mesh_out, mesh_out_std, fit_dd_values = mesh_fit(param_slider, id_type, fit_dd, fit_num,
                                                                          param_center, [invar], [invar1_log],
                                                                          outvar, fit_sampling, add_noise_var)
                for i in range(len(fit_dd_values)):
                    fig.add_trace(go.Scatter(
                        x=mesh_in[i][invars.index(invar)],
                        y=mesh_out[i],
                        mode='lines',
                        name=f'fit: {fit_dd}={fit_dd_values[i]:.1e}',
                        line_color=colormap(indata[fit_dd].min(), indata[fit_dd].max(), fit_dd_values[i]),
                        marker_line=dict(coloraxis="coloraxis2"),
                    ))
                    fig.add_trace(go.Scatter(
                        x=np.hstack((mesh_in[i][invars.index(invar)], mesh_in[i][invars.index(invar)][::-1])),
                        y=np.hstack((mesh_out[i] + fit_conf * mesh_out_std[i], mesh_out[i][::-1] - fit_conf * mesh_out_std[i][::-1])),
                        showlegend=False,
                        fill='toself',
                        line_color=colormap(indata[fit_dd].min(), indata[fit_dd].max(), fit_dd_values[i]),
                        marker_line=dict(coloraxis="coloraxis2"),
                        opacity=fit_opacity,
                    ))
        elif graph_type == '2D':
            fig = go.Figure(
                data=[go.Scatter3d(
                    x=indata[invar][sel_y],
                    y=indata[invar_2][sel_y],
                    z=outdata[outvar][sel_y],
                    mode='markers',
                    name='Data',
                    error_z=dict(type='data', array=outdata[error_dd][sel_y], visible=error_use == ['true'], width= 10)
                )],
                layout=go.Layout(scene=dict(xaxis_title=invar, yaxis_title=invar_2, zaxis_title=outvar))
            )
            if fit_use == ['show'] and invar != invar_2:
                mesh_in, mesh_out, mesh_out_std, fit_dd_values = mesh_fit(param_slider, id_type, fit_dd, fit_num,
                                                                          param_center, [invar, invar_2],
                                                                          [invar1_log, invar2_log], outvar,
                                                                          fit_sampling, add_noise_var)
                for i in range(len(fit_dd_values)):
                    fig.add_trace(go.Surface(
                        x=mesh_in[i][invars.index(invar)].reshape((fit_sampling, fit_sampling)),
                        y=mesh_in[i][invars.index(invar_2)].reshape((fit_sampling, fit_sampling)),
                        z=mesh_out[i].reshape((fit_sampling, fit_sampling)),
                        name=f'fit: {fit_dd}={fit_dd_values[i]:.2f}',
                        surfacecolor=fit_dd_values[i] * np.ones([fit_sampling, fit_sampling])
                            if fit_color == 'multi-fit' else
                            (mesh_in[i][invars.index(color_dd)].reshape((fit_sampling, fit_sampling))
                             if (fit_color == 'marker-color' and color_dd in invars) else
                             mesh_out[i].reshape((fit_sampling, fit_sampling))),
                        opacity=fit_opacity,
                        coloraxis="coloraxis2" if (fit_color == 'multi-fit' or
                            (fit_color == 'output' and (color_dd != outvar and color_dd != 'OUTPUT'))) else "coloraxis",
                        showlegend=True if len(invars) > 2 else False,
                    ))
                    if fit_conf > 0:
                        fig.add_trace(go.Surface(
                            x=mesh_in[i][invars.index(invar)].reshape((fit_sampling, fit_sampling)),
                            y=mesh_in[i][invars.index(invar_2)].reshape((fit_sampling, fit_sampling)),
                            z=mesh_out[i].reshape((fit_sampling, fit_sampling)) + fit_conf * mesh_out_std[i].reshape((fit_sampling, fit_sampling)),
                            showlegend=False,
                            name=f'fit+v: {fit_dd}={fit_dd_values[i]:.2f}',
                            surfacecolor=fit_dd_values[i] * np.ones([fit_sampling, fit_sampling])
                                if fit_color == 'multi-fit' else
                                (mesh_in[i][invars.index(color_dd)].reshape((fit_sampling, fit_sampling))
                                 if (fit_color == 'marker-color' and color_dd in invars) else
                                 mesh_out[i].reshape((fit_sampling, fit_sampling))),
                            opacity=fit_opacity,
                            coloraxis="coloraxis2" if (fit_color == 'multi-fit' or
                                (fit_color == 'output' and (color_dd != outvar and color_dd != 'OUTPUT')))
                                else "coloraxis",
                        ))
                        fig.add_trace(go.Surface(
                            x=mesh_in[i][invars.index(invar)].reshape((fit_sampling, fit_sampling)),
                            y=mesh_in[i][invars.index(invar_2)].reshape((fit_sampling, fit_sampling)),
                            z=mesh_out[i].reshape((fit_sampling, fit_sampling)) - fit_conf * mesh_out_std[i].reshape((fit_sampling, fit_sampling)),
                            showlegend=False,
                            name=f'fit-v: {fit_dd}={fit_dd_values[i]:.2f}',
                            surfacecolor=fit_dd_values[i] * np.ones([fit_sampling, fit_sampling])
                                if fit_color == 'multi-fit' else
                                (mesh_in[i][invars.index(color_dd)].reshape((fit_sampling, fit_sampling))
                                 if (fit_color == 'marker-color' and color_dd in invars) else
                                 mesh_out[i].reshape((fit_sampling, fit_sampling))),
                            opacity=fit_opacity,
                            coloraxis="coloraxis2" if (fit_color == 'multi-fit' or
                                (fit_color == 'output' and (color_dd != outvar and color_dd != 'OUTPUT')))
                                else "coloraxis",
                        ))
                fig.update_layout(coloraxis2=dict(
                    colorbar=dict(title=outvar if fit_color == 'output' else fit_dd),
                    cmin=min(fit_dd_values) if fit_color == 'multi-fit' else None,
                    cmax=max(fit_dd_values) if fit_color == 'multi-fit' else None,
                ))
        elif graph_type == '2D contour':
            mesh_in, mesh_out, mesh_out_std, fit_dd_values = mesh_fit(param_slider, id_type, fit_dd, fit_num,
                                                                      param_center, [invar, invar_2],
                                                                      [invar1_log, invar2_log], outvar,
                                                                      fit_sampling, add_noise_var)
            data_x = mesh_in[0][invars.index(invar)]
            data_y = mesh_in[0][invars.index(invar_2)]
            fig = go.Figure()
            if min(data_x) != max(data_x):
                if min(data_y) != max(data_y):
                    fig.add_trace(go.Scatter(
                        x=indata[invar][sel_y],
                        y=indata[invar_2][sel_y],
                        mode='markers',
                        name='Data',
                    ))
                    fig.add_trace(go.Contour(
                        x=mesh_in[0][invars.index(invar)],
                        y=mesh_in[0][invars.index(invar_2)],
                        z=mesh_out[0],
                        contours_coloring='heatmap',
                        contours_showlabels=True,
                        coloraxis='coloraxis2',
                        name='fit',
                    ))
                    fig.update_xaxes(
                        range=[log10(min(fig.data[1]['x'])), log10(max(fig.data[1]['x']))] if invar1_log == ['log']
                        else [min(fig.data[1]['x']), max(fig.data[1]['x'])])
                    fig.update_yaxes(
                        range=[log10(min(fig.data[1]['y'])), log10(max(fig.data[1]['y']))] if invar2_log == ['log']
                        else [min(fig.data[1]['y']), max(fig.data[1]['y'])])
                    fig.update_layout(xaxis_title=invar,
                                      yaxis_title=invar_2,
                                      coloraxis2=dict(colorbar=dict(title=outvar),
                                                      colorscale='solar',
                                                      cmin=min(fig.data[1]['z']),
                                                      cmax=max(fig.data[1]['z'])))
                else:
                    fig.update_layout(title="y-data is constant, no contour-plot possible")
            else:
                fig.update_layout(title="x-data is constant, no contour-plot possible")
        elif graph_type == '3D':
            fig = go.Figure(
                data=go.Scatter3d(
                    x=indata[invar][sel_y],
                    y=indata[invar_2][sel_y],
                    z=indata[invar_3][sel_y],
                    mode='markers',
                    marker=dict(
                            color=outdata[outvar][sel_y],
                            coloraxis="coloraxis2",
                        ),
                    name='Data',
                ),
                layout=go.Layout(scene=dict(xaxis_title=invar, yaxis_title=invar_2, zaxis_title=invar_3)),
            )
            fig.update_layout(coloraxis2=dict(
                colorbar=dict(title=outvar),
            ))
            if fit_use == ['show'] and len({invar, invar_2, invar_3}) == 3:
                mesh_in, mesh_out, mesh_out_std, fit_dd_values = mesh_fit(param_slider, id_type, fit_dd, fit_num,
                                                                          param_center, [invar, invar_2, invar_3],
                                                                          [invar1_log, invar2_log, invar3_log], outvar,
                                                                          fit_sampling, add_noise_var)
                for i in range(len(fit_dd_values)):
                    fig.add_trace(
                        go.Isosurface(
                            x=mesh_in[i][invars.index(invar)],
                            y=mesh_in[i][invars.index(invar_2)],
                            z=mesh_in[i][invars.index(invar_3)],
                            value=mesh_out[i],
                            surface_count=fit_num,
                            coloraxis="coloraxis2",
                            isomin=mesh_out[i].min() * 1.1,
                            isomax=mesh_out[i].max() * 0.9,
                            caps=dict(x_show=False, y_show=False, z_show=False),
                            opacity=fit_opacity,
                        ),
                    )
        else:
            fig = go.Figure()
        fig.update_layout(legend=dict(xanchor="left", x=0.01))
        # log scale
        log_dict = {'1D': (invar1_log, outvar_log),
                    '2D': (invar1_log, invar2_log, outvar_log),
                    '2D contour': (invar1_log, invar2_log),
                    '3D': (invar1_log, invar2_log, invar3_log),}
        log_list = ['linear' if log is None or len(log) == 0 else log[0] for log in log_dict[graph_type]]
        log_key = ['xaxis', 'yaxis', 'zaxis']
        comb_dict = dict(zip(log_key, [{'type': log} for log in log_list]))
        if len(log_list) < 3 :
            fig.update_layout(**comb_dict)
        else:
            fig.update_scenes(**comb_dict)
        # color
        if color_use == ['true']: # TODO: trigger-detection no new fig just update
            if fit_use == ['show'] and (graph_type=='2D' and (fit_color=='multi-fit' and color_dd==fit_dd)):
                fig.update_traces(
                    marker=dict(
                        coloraxis="coloraxis2",
                        color=indata[color_dd][sel_y] if color_dd in indata.dtype.names else outdata[color_dd][sel_y],
                    ),
                    selector=dict(mode='markers'),
                )
            elif graph_type == '3D':
                fig.update_traces(
                    marker=dict(
                        coloraxis="coloraxis2",
                        color=outdata[outvar][sel_y],
                    ),
                    selector=dict(mode='markers'),
                )
            elif graph_type=='1D':
                fig.update_traces(
                    marker=dict(
                        coloraxis="coloraxis2",
                        color=outdata[outvar][sel_y] if color_dd == 'OUTPUT' else
                        (indata[color_dd][sel_y] if color_dd in indata.dtype.names else outdata[color_dd][sel_y]),
                    ),
                    selector=dict(mode='markers'),
                )
                if color_dd==fit_dd:
                    fig.update_layout(coloraxis2=dict(colorscale='cividis', colorbar=dict(title=fit_dd)))
                elif color_dd == 'OUTPUT':
                    fig.update_layout(coloraxis2=dict(colorscale='plasma', colorbar=dict(title=outvar)))
                else:
                    fig.update_layout(coloraxis2=dict(colorscale='plasma', colorbar=dict(title=color_dd)))
            elif graph_type =='2D contour':
                fig.update_traces(
                    marker=dict(
                        coloraxis="coloraxis",
                        color=outdata[outvar][sel_y] if color_dd == 'OUTPUT' else
                        (indata[color_dd][sel_y] if color_dd in indata.dtype.names else outdata[color_dd][sel_y]),
                    ),
                    selector=dict(mode='markers'),
                )
                if color_dd == outvar or color_dd == 'OUTPUT':
                    fig.update_traces(marker_coloraxis="coloraxis2", selector=dict(mode='markers'))
                else:
                    fig.update_layout(coloraxis=dict(colorbar=dict(title=color_dd, x=1.1),
                                                     colorscale='ice'))
            else:
                fig.update_traces(
                    marker=dict(
                        coloraxis="coloraxis",
                        color=outdata[outvar][sel_y] if color_dd == 'OUTPUT' else
                        (indata[color_dd][sel_y] if color_dd in indata.dtype.names else outdata[color_dd][sel_y]),
                    ),
                    selector=dict(mode='markers'),
                )
                fig.update_layout(coloraxis=dict(
                    colorbar=dict(title=outvar if color_dd == 'OUTPUT' else color_dd, x=1.1),
                    colorscale='viridis',
                ))
        fig.update_layout(height=graph_height)
        return fig


    def mesh_fit(param_slider, id_type, fit_dd, fit_num, param_center, invar_list, invar_log_list, outvar, num_samples, add_noise_var):
        try:  # collecting min/max of slider for variable of multifit
            fit_dd_min, fit_dd_max = param_slider[[i['index'] for i in id_type].index(invars.index(fit_dd))]
        except ValueError:
            fit_dd_min = min(indata[fit_dd])
            fit_dd_max = max(indata[fit_dd])

        if fit_num == 1: # generate list of value of variable of multifit
            fit_dd_values = np.array([(fit_dd_max + fit_dd_min) / 2])
        else:
            fit_dd_values = np.linspace(fit_dd_min, fit_dd_max, fit_num)

        for iteration, fit_dd_value in enumerate(fit_dd_values): # iteration for each fit
            # set fit parameter for all invars as center of range
            fit_params = [(max(indata[var_invar]) + min(indata[var_invar])) / 2 for var_invar in invars]
            # for all invars with filter change fit_param to center defined by filter
            flt_ind_list = [] # list of filter indices
            for i, center_values in enumerate(param_center):
                flt_ind_list.append(id_type[i]['index'])
                fit_params[flt_ind_list[i]] = center_values
            # change param of fit-variable
            fit_params[invars.index(fit_dd)] = fit_dd_value
            # change param for axis invars
            for i, ax_in in enumerate(invar_list):
                if invars.index(ax_in) in flt_ind_list:
                    ax_min, ax_max = param_slider[flt_ind_list.index(invars.index(ax_in))]
                else:
                    ax_min = min(indata[ax_in])
                    ax_max = max(indata[ax_in])
                if invar_log_list[i] == ['log']:
                    fit_params[invars.index(ax_in)] = np.logspace(log10(ax_min), log10(ax_max), num_samples)
                else:
                    fit_params[invars.index(ax_in)] = np.linspace(ax_min, ax_max, num_samples)
            grid = np.meshgrid(*fit_params) # generate grid
            x_pred = np.vstack([g.flatten() for g in grid]).T  # extract vector for predict
            fit_data, fit_var = sur.predict(x_pred, add_noise_var == ['add']) # generate fit data an variance
            # generated data
            new_mesh_in = np.array([[grid[invars.index(invar)].flatten() for invar in invars]])
            new_mesh_out = np.array([fit_data[:, outvars.index(outvar)]])
            new_mesh_out_std = np.array([np.sqrt(fit_var[:, 0])])
            # stack data together
            if iteration == 0:
                mesh_in = new_mesh_in
                mesh_out = new_mesh_out
                mesh_out_std = new_mesh_out_std
            else:
                mesh_in = np.vstack((mesh_in, new_mesh_in))
                mesh_out = np.vstack((mesh_out, new_mesh_out))
                mesh_out_std = np.vstack((mesh_out_std, new_mesh_out_std))
        return mesh_in, mesh_out, mesh_out_std, fit_dd_values


    def create_slider(dd_value):
        ind = invars.index(dd_value)
        slider_min = indata[dd_value].min()
        slider_max = indata[dd_value].max()
        step_exponent = -3
        new_slider = dcc.RangeSlider(
            id={'type': 'param-slider', 'index': ind},
            step=10 ** step_exponent,
            min=slider_min,
            max=slider_max,
            value=[slider_min, slider_max],
            marks={slider_min: str(round(slider_min, -step_exponent)),
                   slider_max: str(round(slider_max, -step_exponent))},
        )
        return new_slider


    def colormap(cmin, cmax, c):
        if cmin == cmax:
            c_scal = 0.5
        else:
            c_scal = (c-cmin)/(cmax-cmin)
        return color2hex(colormaps.cividis(c_scal))

    return app



import GPy
import numpy as np
import matplotlib.pyplot as plt

f = lambda x: np.sin(x)  # our function
fd = lambda x: np.cos(x)  # its derivative

d = 1  # Dimension
n = 10  # Number of observations
n_der = 10  # Number of derivative observations
n_star = 100  # Number of prediction points

sigma_n = 1e-2  # Noise of observations
sigma_n_der = 1e-2  # Noise of derivative observations
x = np.linspace(1, 10, n).reshape((-1, d))  # training points
y = f(x) + np.array(sigma_n * np.random.normal(0, 1, (n, d)))  # training outputs
x_der = np.linspace(2, 8, n_der).reshape((-1, d))  # derivative training points
y_der = fd(x_der) + np.array(sigma_n_der * np.random.normal(0, 1, (n_der, d)))  # derivative training outputs
x_star = np.linspace(0, 11, n_star).reshape((-1, d))  # test points
y_star_true = f(x_star)  # the real value of our test points according to f
y_der_star_true = fd(x_star)  # the real value of our test points according to fd

l = 0.2  # Lengthscale
sigma_f = 1  # Length of the x-axis
k = GPy.kern.RBF(input_dim=d, lengthscale=l,
                 variance=sigma_f ** 2)  # we choose the squared exponential function as a kernel
k_der = GPy.kern.DiffKern(k, 0)  # the derivative of our kernel
gauss = GPy.likelihoods.Gaussian(variance=sigma_n ** 2)  # Gaussian likelihood
gauss_der = GPy.likelihoods.Gaussian(variance=sigma_n_der ** 2)  # Derivative Gaussian likelihood

m = GPy.models.MultioutputGP(X_list=[x, x_der], Y_list=[y, y_der], kernel_list=[k, k_der],
                             likelihood_list=[gauss, gauss_der])  # We are building our model " m "
print(m)  # we are printing our model


def plot_gp_vs_real(m, x, yreal, size_inputs, title, fixed_input=1, xlim=[0, 11], ylim=[-1.5, 3]):
    fig, ax = plt.subplots()
    ax.set_title(title)
    plt.plot(x, yreal, "r", label=' Real function ')
    rows = slice(0, size_inputs[0]) if fixed_input == 0 else slice(size_inputs[0], size_inputs[0] + size_inputs[1])
    m.plot(fixed_inputs=[(1, fixed_input)], which_data_rows=rows, xlim=xlim, ylim=ylim, ax=ax)


# Plot the model , the syntax is same as for multi - output models :
plot_gp_vs_real(m, x_star, y_der_star_true, [x.shape[0], x_der.shape[0]], title=' Latent function derivatives ',
                fixed_input=1, xlim=[0, 8], ylim=[-6.5, 6.5])
plot_gp_vs_real(m, x_star, y_star_true, [x.shape[0], x_der.shape[0]], title=' Latent function ', fixed_input=0,
                xlim=[0, 8], ylim=[-2.1, 2.1])

m.optimize()  # Optimisation of the parameters
print(m)  # Showing the model

plot_gp_vs_real(m, x_star, y_der_star_true, [x.shape[0], x_der.shape[0]], title=' Latent function derivatives ',
                fixed_input=1, xlim=[0, 8], ylim=[-1.1, 1.1])
plot_gp_vs_real(m, x_star, y_star_true, [x.shape[0], x_der.shape[0]], title=' Latent function ', fixed_input=0,
                xlim=[0, 8], ylim=[-1.1, 1.1])
plt.show()



"""Halton low discrepancy sequence.

This snippet implements the Halton sequence following the generalization of
a sequence of *Van der Corput* in n-dimensions.

---------------------------

MIT License

Copyright (c) 2017 Pamphile Tupui ROY

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import numpy as np


def primes_from_2_to(n):
    """Prime number from 2 to n.

    From `StackOverflow <https://stackoverflow.com/questions/2068372>`_.

    :param int n: sup bound with ``n >= 6``.
    :return: primes in 2 <= p < n.
    :rtype: list
    """
    sieve = np.ones(n // 3 + (n % 6 == 2), dtype=bool)
    for i in range(1, int(n ** 0.5) // 3 + 1):
        if sieve[i]:
            k = 3 * i + 1 | 1
            sieve[k * k // 3::2 * k] = False
            sieve[k * (k - 2 * (i & 1) + 4) // 3::2 * k] = False
    return np.r_[2, 3, ((3 * np.nonzero(sieve)[0][1:] + 1) | 1)]


def van_der_corput(n_sample, base=2):
    """Van der Corput sequence.

    :param int n_sample: number of element of the sequence.
    :param int base: base of the sequence.
    :return: sequence of Van der Corput.
    :rtype: list (n_samples,)
    """
    sequence = []
    for i in range(n_sample):
        n_th_number, denom = 0., 1.
        while i > 0:
            i, remainder = divmod(i, base)
            denom *= base
            n_th_number += remainder / denom
        sequence.append(n_th_number)

    return sequence


def halton(n_sample, dim):
    """Halton sequence.

    :param int n_sample: number of samples.
    :param int dim: dimension
    :return: sequence of Halton.
    :rtype: array_like (n_samples, n_features)
    """
    big_number = 10
    while 'Not enought primes':
        base = primes_from_2_to(big_number)[:dim]
        if len(base) == dim:
            break
        big_number += 1000

    # Generate a sample using a Van der Corput sequence per dimension.
    sample = [van_der_corput(n_sample + 1, dim) for dim in base]
    sample = np.stack(sample, axis=-1)[1:]

    return sample



"""
Abstract base class for all surrogate models.

Class structure:
- Surrogate
    - GaussianProcess
        - GPSurrogate (Custom)
        - GPySurrogate (GPy)
        - SklearnGPSurrogate (Sklearn)
    - ANN
        - ANNSurrogate (Pytorch)
        - Autoencoder (Pytorch)
    - LinearRegression
        - Linear regression surrogates (Work in progress)
"""

from abc import ABC, abstractmethod
import numpy as np


class Surrogate(ABC):
    """Base class for all surrogate models.

    Attributes:
        trained (bool): Flag that indicates if the model is already trained and ready to make predictions.
        fixed_sigma_n (bool): Indicates if the data noise should be optimized or not.
        Xtrain (ndarray): Input training points.
        ytrain (ndarray): Observed output data.
            Vector output is supported for independent variables only.
        ndim (int): Dimension of input data.
        output_ndim (int): Dimension of output data.
        multi_output (bool): True, if more than one output variable is defined in the
            config file. If False, excess output dimensions are used as independent supporting points.
        encoder (list of profit.sur.encoders.Encoder): For now, inputs of kind 'LogUniform' are encoded with 'log10'
            and all input and output data is normalized. Can be modified in the config file using the format
            e.g. [['log10', [0], False], ['normalization', [0, 1], False]].

    Default parameters:
        surrogate: GPy
        save: ./model_{surrogate_label}.hdf5
        load: False
        fixed_sigma_n: False
        encoder: [['log10', [log_input_cols], False], ['normalization', [input_cols], False],
            ['normalization', [output_cols], True]]
    """

    _surrogates = {}  # All surrogates are registered here
    _defaults = {'surrogate': 'GPy',  # Default surrogate configuration parameters
                 'save': './model.hdf5',
                 'load': False,
                 'fixed_sigma_n': False,
                 'encoder': []}

    def __init__(self):
        self.trained = False
        self.fixed_sigma_n = False
        self.multi_output = False  # Inferred from output variables in config.

        self.Xtrain = None
        self.ytrain = None
        self.ndim = None  # TODO: Consistency between len(base_config['input']) and self.Xtrain.shape[-1]
        self.output_ndim = 1

        self.encoder = []

    def encode_training_data(self):
        """Encodes the input and output training data.
        """
        for enc in self.encoder:
            if enc.output:
                self.ytrain = enc.encode(self.ytrain)
            else:
                self.Xtrain = enc.encode(self.Xtrain)

    def decode_training_data(self):
        """Applies the decoding function of the encoder in reverse order on the input and output training data.
        """
        for enc in self.encoder[::-1]:
            if enc.output:
                self.ytrain = enc.decode(self.ytrain)
            else:
                self.Xtrain = enc.decode(self.Xtrain)

    def encode_predict_data(self, x):
        """Transforms the input prediction points according to the encoder used for training.

        Parameters:
            x (ndarray): Prediction input points.

        Returns:
            ndarray: Encoded and normalized prediction points.
        """
        for enc in self.encoder:
            if not enc.output:
                if enc.label == 'normalization':
                    x[:, enc.columns] = x[:, enc.columns] / enc.variables['xmax']
                else:
                    x = enc.encode(x)
        return x

    def decode_predict_data(self, ym, yv):
        """Rescales and then back-transforms the predicted output.

        Parameters:
            ym (ndarray): Predictive output.
            yv (ndarray): Variance of predicted output.

        Returns:
            tuple: a tuple containing:
                - ym (ndarray) Rescaled and decoded output values at the test input points.
                - yv (ndarray): Rescaled predictive variance.
        """

        for enc in self.encoder[::-1]:
            if enc.output:
                if enc.label == 'normalization':
                    ym = ym * enc.variables['xmax']
                    yv = yv * enc.variables['xmax'] ** 2
                else:
                    ym = enc.decode(ym)
        return ym, yv

    @abstractmethod
    def train(self, X, y, fixed_sigma_n=False, multi_output=False):
        r"""Trains the surrogate on input points X and model outputs y.

        Depending on the surrogate, the signature can vary.

        Parameters:
            X (ndarray): Input training points.
            y (ndarray): Observed output data.
            fixed_sigma_n (bool): Whether the noise $\sigma_n$ is fixed during optimization.
            multi_output (bool): Whether a multi output model should be used for fitting.
        """
        pass

    @abstractmethod
    def predict(self, Xpred, add_data_variance=True):
        r"""Predicts model output y for input Xpred based on surrogate.

        Parameters:
            Xpred (ndarray/list): Input points for prediction.
            add_data_variance (bool): Adds the data noise $\sigma_n^2$ to the prediction variance.
                This is especially useful for plotting.

        Returns:
            tuple: a tuple containing:
                - ymean (ndarray) Predicted output values at the test input points.
                - yvar (ndarray): Generally the uncertainty of the fit. For Gaussian Processes this is
                the diagonal of the posterior covariance matrix.
        """
        pass

    @abstractmethod
    def save_model(self, path):
        """Saves the surrogate to a file. The file format can vary between surrogates.
        As default, the surrogate is saved to 'base_dir/model_{surrogate_label}.hdf5'.

        Parameters:
            path (str): Path including the file name, where the model should be saved.
        """
        pass

    @classmethod
    @abstractmethod
    def load_model(cls, path):
        """Loads a saved surrogate from a file. The file format can vary between surrogates.

        Identifies the surrogate by its class label in the file name.

        Parameters:
            path (str): Path including the file name, from where the model should be loaded.

        Returns:
            profit.sur.Surrogate: Instantiated surrogate model.
        """
        label = next(filter(lambda l: l in path, cls._surrogates), cls._defaults['surrogate'])
        return cls[label].load_model(path)

    @classmethod
    @abstractmethod
    def from_config(cls, config, base_config):
        """Instantiates a surrogate based on the parameters given in the configuration file and delegates to child.

        Parameters:
            config (dict): Only the 'fit' part of the base_config.
            base_config (dict): The whole configuration parameters.
        """
        from .encoders import Encoder

        child = cls[config['surrogate']]

        if config.get('load'):
            child_instance = child.load_model(config['load'])
        else:
            child_instance = child.from_config(config, base_config)
            # Set global attributes
            child_instance.ndim = len(base_config['input'])
            child_instance.output_ndim = len(base_config['output'])
            child_instance.multi_output = len(base_config['output']) > 1
            child_instance.fixed_sigma_n = config['fixed_sigma_n']
        child_instance.encoder = [Encoder(func, cols, out) for func, cols, out in config['encoder']]
        return child_instance

    @classmethod
    def handle_config(cls, config, base_config):
        """Fills the configuration parameters with defaults, if not existent, and delegates to child.

        If saving or loading is enabled, the class label is included in the filename to identify the
        surrogate class. Relative paths are referenced to the base directory.

        Parameters:
            config (dict): Only the 'fit' part of the base_config.
            base_config (dict): The whole configuration parameters.
        """
        for key, default in cls._defaults.items():
            if key not in config:
                config[key] = default
        if not config['encoder']:
            in_dims = [idx for idx, value in enumerate(base_config['input'].values()) if value['kind'] != 'constant']
            out_dims = list(range(len(base_config['output'].keys())))
            log_input = [idx for idx, value in enumerate(base_config['input'].values())
                         if value['kind'] == 'LogUniform']
            config['encoder'] = [['log10', log_input, False],
                                 ['normalization', in_dims, False],
                                 ['normalization', out_dims, True]]

        for mode in ('save', 'load'):
            if config.get(mode):
                from os.path import abspath, join
                config[mode] = abspath(join(base_config['base_dir'], config[mode]))
                if config['surrogate'] not in config[mode]:
                    filepath = config[mode].rsplit('.', 1)
                    config[mode] = ''.join(filepath[:-1]) + f'_{config["surrogate"]}.' + filepath[-1]
        if config.get('load'):
            config['save'] = False
        Surrogate[config['surrogate']].handle_subconfig(config, base_config)

    @classmethod
    @abstractmethod
    def handle_subconfig(cls, config, base_config):
        """Fills configuration parameters with child's defaults.

        Parameters:
            config (dict): Only the 'fit' part of the base_config.
            base_config (dict): The whole configuration parameters.
        """
        pass

    @classmethod
    def register(cls, label):
        """Decorator to register new surrogate classes."""
        def decorator(surrogate):
            if label in cls._surrogates:
                raise KeyError(f'registering duplicate label {label} for surrogate.')
            cls._surrogates[label] = surrogate
            return surrogate
        return decorator

    def __class_getitem__(cls, item):
        """Returns the child surrogate."""
        return cls._surrogates[item]

    @classmethod
    def get_label(cls):
        """Returns the string label of a surrogate class object."""
        for label, item in cls._surrogates.items():
            if item == cls:
                return label
        raise NotImplementedError("Class {} is not implemented.".format(cls))

    def plot(self, Xpred=None, independent=None, show=False, ref=None, add_data_variance=True, axes=None):
        r"""Simple plotting for dimensions <= 2.

        Fore more sophisticated plots use the command 'profit ui'.

        Parameters:
            Xpred (ndarray): Prediction points where the fit is plotted. If None, it is inferred from the
                training points.
            independent (dict): Dictionary of independent variables from config.
            show (bool): If the figure should be shown directly.
            ref (ndarray): Reference function which is fitted.
            add_data_variance (bool): Adds the data noise $\sigma_n^2$ to the prediction variance.
            axes (matplotlib.pyplot.axes): Axes object to insert the plot into. If None, a new figure is created.
        """

        import matplotlib.pyplot as plt
        if Xpred is None:
            Xpred = self.default_Xpred()
        ypred, yvarpred = self.predict(Xpred, add_data_variance=add_data_variance)
        ystd_pred = np.sqrt(yvarpred)
        if independent:
            # 2D with one input parameter and one independent variable.
            if self.ndim == 1 and ypred.ndim == 2:
                ax = axes or plt.axes(projection='3d')
                xind = np.hstack([v['range'] for k, v in independent.items()])
                xtgrid = np.meshgrid(*[xind, self.Xtrain])
                xgrid = np.meshgrid(*[xind, Xpred])
                for i in range(self.Xtrain.shape[0]):
                    ax.plot(xtgrid[0][i], xtgrid[1][i], self.ytrain[i], color='blue', linewidth=2)
                ax.plot_surface(xgrid[0], xgrid[1], ypred, color='red', alpha=0.8)
                ax.plot_surface(xgrid[0], xgrid[1], ypred + 2 * ystd_pred, color='grey', alpha=0.6)
                ax.plot_surface(xgrid[0], xgrid[1], ypred - 2 * ystd_pred, color='grey', alpha=0.6)
            else:
                raise NotImplementedError("Plotting is only implemented for dimensions <= 2. Use profit ui instead.")
        else:
            if self.ndim == 1 and ypred.shape[-1] == 1:
                # Only one input parameter to plot.
                ax = axes or plt.axes()
                if ref:
                    ax.plot(Xpred, ref(Xpred), color='red')
                ax.plot(Xpred, ypred)
                ax.scatter(self.Xtrain, self.ytrain, marker='x', s=50, c='k')
                ax.fill_between(Xpred.flatten(),
                                ypred.flatten() + 2 * ystd_pred.flatten(), ypred.flatten() - 2 * ystd_pred.flatten(),
                                color='grey', alpha=0.6)
            elif self.ndim == 2 and ypred.shape[-1] == 1:
                # Two fitted input variables.
                ax = axes or plt.axes(projection='3d')
                ypred = ypred.flatten()
                ystd_pred = ystd_pred.flatten()
                ax.scatter(self.Xtrain[:, 0], self.Xtrain[:, 1], self.ytrain, color='red', alpha=0.8)
                ax.plot_trisurf(Xpred[:, 0], Xpred[:, 1], ypred, color='red', alpha=0.8)
                ax.plot_trisurf(Xpred[:, 0], Xpred[:, 1], ypred + 2 * ystd_pred, color='grey', alpha=0.6)
                ax.plot_trisurf(Xpred[:, 0], Xpred[:, 1], ypred - 2 * ystd_pred, color='grey', alpha=0.6)
            elif self.ndim == 1 and self.output_ndim == 2:
                # One input variable and two outputs
                ax = axes or plt.axes()
                for d in range(self.output_ndim):
                    yp = ypred[:, d]
                    ystd_p = ystd_pred[:, d]
                    ax.scatter(self.Xtrain, self.ytrain[:, d], alpha=0.8)
                    ax.plot(Xpred, yp)
                    ax.fill_between(Xpred.flatten(), yp + 2 * ystd_p, yp - 2 * ystd_p, alpha=0.6)
            else:
                raise NotImplementedError("Plotting is only implemented for dimension <= 2. Use profit ui instead.")

        if show:
            plt.show()

    def default_Xpred(self):
        """Infer prediction values from training points in each dimension.

        Currently a dense grid is created. This becomes inefficient for > 3 dimensions.

        Returns:
            ndarray: Prediction points.
        """

        if self.ndim <= 3:
            minval = self.Xtrain.min(axis=0)
            maxval = self.Xtrain.max(axis=0)
            npoints = [50] * len(minval)
            xpred = [np.linspace(minv, maxv, n) for minv, maxv, n in zip(minval, maxval, npoints)]
            return np.hstack([xi.flatten().reshape(-1, 1) for xi in np.meshgrid(*xpred)])
        else:
            raise RuntimeError("Require x for prediction in > 3 dimensions!")



import numpy as np
import GPy

k0 = GPy.kern.RBF(1, active_dims=0)
k1 = GPy.kern.RBF(1, active_dims=1)

k0_der = GPy.kern.DiffKern(k0, 0)

# Extended class for product kernel,
# can be merged with Prod class when finished

class ProdExtended(GPy.kern.Prod):    
    def dK_dX(self, X, X2, dimX):
        # Product rule to kprod = k0*k1
        # d(kprod/dx0) = dk0/dx0*k1
        # d(kprod/dx1) = k0*dk1/dx1
        other = self.parts[:] 
        All = other[:]
        n = len(All)
        nX = X.shape[0]
        nX2 = X2.shape[0]
        diffpart = other.pop(dimX)
        K_all = np.zeros((n,nX,nX2))
        for k in range(n):
            if (k!=dimX):
                K_all[k,:,:] = All[k].K(X,X2)
            else:
                K_all[k,:,:] = diffpart.dK_dX(X,X2,dimX)
        return np.prod(K_all,0)
        
    def dK_dX2(self,X,X2,dimX2):
        return -self.dK_dX(X,X2,dimX2)

    def dK2_dXdX2(self, X, X2, dimX, dimX2):
        # Product rule to kprod = k0*k1
        # d²(kprod/dx0²) = d²k0/dx0²*k1
        # d²(kprod/dx1²) = k0*d²k1/dx1²
        # d²(kprod/dx0dx1) = dk0/dx0*dk1/dx1
        # d²(kprod/dx1dx0) = dk0/dx0*dk1/dx1
        other = self.parts[:] 
        All = other[:]
        n = len(All)
        nX = X.shape[0]
        nX2 = X2.shape[0]
        if (dimX==dimX2):
            diffpart = other.pop(dimX)
            K_all = np.zeros((n,nX,nX2))
            for k in range(n):
                if (k!=dimX):
                    K_all[k,:,:] = All[k].K(X,X2)
                else:
                    K_all[k,:,:] = diffpart.dK2_dXdX2(X,X2,dimX,dimX)
        else:
            diffpart1 = other.pop(dimX)
            diffpart2 = All[dimX2]
            K_all = np.zeros((n,nX,nX2))
            for k in range(n):
                if (k==dimX):
                    K_all[k,:,:] = diffpart1.dK_dX(X,X2,dimX)
                elif (k==dimX2):
                    K_all[k,:,:] = diffpart2.dK_dX2(X,X2,dimX2)
                else:
                    K_all[k,:,:] = All[k].K(X,X2)
        return np.prod(K_all,0)




#%%
import numpy as np
import plotly.express as px
import pandas as pd
import tensorflow as tf
from tensorflow import keras

# data = pd.read_table(
#     r'C:\Users\chral\Dropbox\ipp\paper_algae\mc_out.dat', sep='\s+')
data = pd.read_table(
    '/home/calbert/Dropbox/ipp/paper_algae/mc_out.dat', sep='\s+')

names = {
    'mu_0': 'k_alg_growth_max',
    'f_si': 'frac_si_alg_1',
    'lambda_S': 'k_att_shade',
    'K_light': 'k_light_sm',
    'sigma_0': 'k_alg_loss',
    'a': 'coeff_d_loss_2'
}
# %%
Ndim = 6
nskip = 1000
indata = data[['k_alg_loss', 'frac_si_alg_1']][::nskip].values
indata = (indata - np.min(indata,0))/np.max(indata - np.min(indata,0),0)
px.scatter(x=indata[:,0], y=indata[:,1])
# %%
# Sequential api
model2 = keras.Sequential()
model2.add(keras.layers.Dense(16, input_dim=2, activation=tf.nn.tanh))
model2.add(keras.layers.Dense(256, activation=tf.nn.tanh))
model2.add(keras.layers.Dense(16, activation=tf.nn.tanh))
model2.add(keras.layers.Dense(1, activation=tf.nn.tanh))
model2.add(keras.layers.Dense(16, activation=tf.nn.tanh))
model2.add(keras.layers.Dense(256, activation=tf.nn.tanh))
model2.add(keras.layers.Dense(16, activation=tf.nn.tanh))
model2.add(keras.layers.Dense(2, activation=tf.nn.sigmoid))

model2.compile(optimizer=tf.optimizers.Adam(), loss='mse')
#%%
dataset = tf.data.Dataset.from_tensor_slices((indata, indata))
for feat, targ in dataset.take(5):
  print ('Features: {}, Target: {}'.format(feat, targ))

train_dataset = dataset.shuffle(10000).batch(1)
#%%

#tb_callback = keras.callbacks.TensorBoard(log_dir='./tb', histogram_freq=0, write_graph=True, write_images=True)

history = model2.fit(train_dataset, epochs = 128)#, callbacks=[tb_callback])
# %%
output_eval_train = model2.predict(indata)
# %%

fig = px.scatter(x=indata[:,0], y=indata[:,1])
fig.add_scatter(x=output_eval_train[:,0], y=output_eval_train[:,1], mode='markers')
# %%



""" mockup 1D integrated directly into the worker

the custom worker directly replaces the main function, disabling all pre and post steps
thereby all configuration for pre and post has no effect (actually it might have an effect on the runner)
the interface can be specified normally with the config file as usual
"""

from profit.run import Worker
import numpy as np


def f(u):
    return np.cos(10 * u) + u


class MockupWorker(Worker):
    def main(self):
        self.interface.output['f'] = f(self.interface.input['u'])
        self.interface.done()


if __name__ == '__main__':
    worker = MockupWorker.from_env()
    worker.main()



import numpy as np
import scipy as sp
from profit.sur.backend.gp_functions_old import invert_cholesky, solve_cholesky

A = np.array([[2,1],[1,2]])
print(np.linalg.inv(A))
L = sp.linalg.cholesky(A)
L_np = np.linalg.cholesky(A)
print(L)
print(L_np)

print()
print(invert_cholesky(L))
print(invert_cholesky(L_np))

b = np.array([1,-2])
print()
print(np.linalg.solve(A,b))
print(solve_cholesky(L,b))
print(solve_cholesky(L_np,b))




import numpy as np
from profit.sur.backend.gp_functions import solve_cholesky
from profit.sur.backend.python_kernels import RBF

def f(x):
    return np.sin(x)

nx = 100
train_every = 10
x = np.linspace(0, 5, nx).reshape([nx, 1])
y = f(x)
xtrain = x[::train_every]
ytrain = f(xtrain)
nxtrain = len(xtrain)

a = np.array([1.0, 1.0])

def test_gp_1D():

    Ky = RBF(xtrain, xtrain, *a)

    assert np.array_equal(Ky, Ky.T)

    L = np.linalg.cholesky(Ky)
    alpha = solve_cholesky(L, ytrain)
    alpha2 = np.linalg.solve(Ky, ytrain)
    Kyinv = np.linalg.inv(Ky)
    alpha3 = Kyinv.dot(ytrain)
    assert np.allclose(alpha2, alpha, rtol=1e-10, atol=1e-10)
    assert np.allclose(alpha3, alpha, rtol=1e-10, atol=1e-10)

    KstarT = RBF(x, xtrain, *a)
    ypred = KstarT.dot(alpha)

    assert np.allclose(ytrain, ypred[::train_every], rtol=1e-14, atol=1e-14)
    assert np.allclose(y, ypred, rtol=1e-2, atol=1e-2)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Aug 28 02:08:02 2020

@author: manal
"""

import numpy as np
import GPy
from GPy.kern.src.stationary import Stationary


class Sinus_prod(Stationary):
    """
    Sinus kernel: 
        Product of 1D Sinus kernels

    .. math::

        &k(x,x')_i = \sigma^2 \prod_{j=1}^{dimension} \sin(x_{i,j}-x_{i,j}') 
        
        &x,x' \in \mathcal{M}_{n,dimension}
        
        &k \in \mathcal{M}_{n,n}

    """
    def __init__(self, input_dim, variance=1., lengthscale=None, ARD=False, active_dims=None, name='Sinus_prod'):
        super(Sinus_prod, self).__init__(input_dim, variance, lengthscale, ARD, active_dims, name)

    def K_of_r(self, dist):
        n = dist.shape[2]
        p = 1
        # l = self.lengthscale
        for k in range(n):
            p*= np.sin(dist[:,:,k])#/l)
        return self.variance * p
    
    def K(self, X, X2):
        dist = X[:,None,:]-X2[None,:,:]
        return self.K_of_r(dist)

    def dK_dr(self,dist,dimX):
        n = dist.shape[2]
        m = dist.shape[0]
        # l = self.lengthscale
        dK = np.zeros((m,m,n))
        for i in range(n):
            dK[:,:,i]= np.sin(dist[:,:,i])#/l)
        dK[:,:,dimX] = np.cos(dist[:,:,dimX])#/l)
        return self.variance * np.prod(dK,2)#/l
    
    def dK_dX(self, X, X2, dimX):
        dist = X[:,None,:]-X2[None,:,:]
        dK_dr = self.dK_dr(dist,dimX)
        return dK_dr
    
    def dK_dX2(self,X,X2,dimX2):
        return -self.dK_dX(X,X2, dimX2)
    
    def dK2_dXdX2(self, X, X2, dimX, dimX2):
        dist = X[:,None,:]-X2[None,:,:]
        K = self.K_of_r(dist)
        n = dist.shape[2]
        m = dist.shape[0]
        # l = self.lengthscale
        dK = np.zeros((m,m,n))
        for i in range(n):
            dK[:,:,i]= np.sin(dist[:,:,i])#/l)
        dK[:,:,dimX] = np.cos(dist[:,:,dimX])#/l)
        dK[:,:,dimX2] = np.cos(dist[:,:,dimX2])#/l)
        return ((dimX==dimX2)*K - (dimX!=dimX2)*np.prod(dK,2))#/(l**2)
    



import GPy
import numpy as np
from time import time
from scipy.optimize import minimize
from profit.util.halton import halton
from profit.sur.backend.gp_functions_old import nll_chol

ntrain = 32
xtrain = halton(ntrain, 2)
K = np.empty((ntrain, ntrain), order='F')
dK = np.empty((ntrain, ntrain), order='F')
l0 = np.array([0.5, 0.5])
sig2f0 = 1.0
sig2n0 = 1e-8
hyp = np.hstack([1.0/l0**2, [sig2f0, sig2n0/sig2f0]])

ytrain = np.sin(2*xtrain[:,0]) + np.cos(3*xtrain[:,1])

loghyp0 = np.log10(hyp)

def cost(loghyp):
    h = 10**loghyp
    val, jac = nll_chol(h, xtrain, ytrain, K, dK)
    return val, h*np.log(10.0)*jac

tic = time()
res = minimize(
    cost, loghyp0, method='L-BFGS-B', jac=True, bounds=(
        (-2, 2), (-2, 2), (-2, 2), (-7, 0)
    )
)
print(f'Time optimize proFit: {time() - tic} s')
print(res)

l = np.sqrt(1.0/10.0**res.x[:-2])
sig2f = 10.0**res.x[-2]
sig2n = sig2f*10.0**res.x[-1]
print(l)
print(sig2f)
print(sig2n)

kernel = GPy.kern.RBF(input_dim=2, variance=sig2f0, lengthscale=l0, ARD=True)
m = GPy.models.GPRegression(
    xtrain, ytrain.reshape(-1, 1), kernel, noise_var=sig2n0
)

tic = time()
m.optimize(messages=False)
print(f'Time optimize GPy: {time() - tic} s')

hyp_test = np.array([
    1.0/m.kern.lengthscale[0]**2,
    1.0/m.kern.lengthscale[1]**2,
    m.kern.variance[0],
    (m.Gaussian_noise.variance[0] + 1e-8)/m.kern.variance[0]
])

print(cost(np.log10(hyp_test))[0])

m.log_likelihood()
m



""" Custom postprocessor """
from profit.run import Postprocessor

import numpy as np


@Postprocessor.register('simulation')
class SimulationPostprocessor(Postprocessor):
    def post(self, data):
        data['f'] = np.loadtxt('result.out')

    @classmethod
    def handle_config(cls, config, base_config):
        pass



import numpy as np
import matplotlib.pyplot as plt
import profit as pf
import GPy

xtrain = pf.util.quasirand(20, 2)
n1 = 20
n2 = 20
x1, x2 = np.meshgrid(np.linspace(0,1,n1), np.linspace(0,1,n2))
xtest = np.vstack([x1.flatten(), x2.flatten()]).T

def f(x):
    return np.sin(4*(x[:,0] + x[:,1]))

ytrain = f(xtrain).reshape(-1,1)
ytest = f(xtest)
plt.figure()
plt.contourf(x1, x2, ytest.reshape(n2, n1))

k = GPy.kern.RBF(2, .5, [0.1, 0.1], ARD=True)
m = GPy.models.GPRegression(xtrain, ytrain, k)
res = m.optimize()
m.plot()
m.kern.lengthscale

#%%
GPy.examples.dimensionality_reduction.bcgplvm_linear_stick()

# %%



#%%
import numpy as np
import matplotlib.pyplot as plt
import torch
import gpytorch

ntrain = 256
lengthscale = 0.25
outputscale = 4.0**2
noise = 1e-4
fast = True

def f(x): return 10.0*((x[0]**2-x[0])*np.cos(24*x[1]) + 0.1*np.cos(24*x[0]))   # original model

#-- from https://laszukdawid.com/2017/02/04/halton-sequence-in-python/
def next_prime():
    def is_prime(num):
        "Checks if num is a prime value"
        for i in range(2,int(num**0.5)+1):
            if(num % i)==0: return False
        return True
 
    prime = 3
    while(1):
        if is_prime(prime):
            yield prime
        prime += 2

def vdc(n, base=2):
    vdc, denom = 0, 1
    while n:
        denom *= base
        n, remainder = divmod(n, base)
        vdc += remainder/float(denom)
    return vdc

def halton_sequence(size, dim):
    seq = []
    primeGen = next_prime()
    next(primeGen)
    for d in range(dim):
        base = next(primeGen)
        seq.append([vdc(i, base) for i in range(size)])
    return np.array(seq).T


#--

xtrain = halton_sequence(ntrain, 2)
ytrain = f(xtrain.T)
U, V = np.meshgrid(np.linspace(0,1,20), np.linspace(0,1,20))
Yref = f([U, V])

Yref = f([U, V])

plt.figure()
plt.contourf(U, V, Yref, cmap='coolwarm')
plt.plot(xtrain[:,0], xtrain[:,1], 'xk')
plt.plot([0,1],[0.5,0.5],'k')
plt.xlabel('$x_1$')
plt.ylabel('$x_2$')

# %%
with gpytorch.settings.fast_computations(fast), \
    gpytorch.settings.max_preconditioner_size(512), \
    gpytorch.settings.min_preconditioning_size(256):
    bounds = np.array([[0,1]])
    train_x = torch.from_numpy(xtrain).float()
    train_y = torch.from_numpy(ytrain).float()
    # We will use the simplest form of GP model, exact inference
    class ExactGPModel(gpytorch.models.ExactGP):
        def __init__(self, train_x, train_y, likelihood):
            super(ExactGPModel, self).__init__(train_x, train_y, likelihood)
            self.mean_module = gpytorch.means.ZeroMean()
            self.covar_module = gpytorch.kernels.ScaleKernel(gpytorch.kernels.MaternKernel())

        def forward(self, x):
            mean_x = self.mean_module(x)
            covar_x = self.covar_module(x)
            return gpytorch.distributions.MultivariateNormal(mean_x, covar_x)

    # initialize likelihood and model
    likelihood = gpytorch.likelihoods.GaussianLikelihood()
    model = ExactGPModel(train_x, train_y, likelihood)


    # Find optimal model hyperparameters
    model.covar_module.base_kernel.initialize(lengthscale=lengthscale)
    model.covar_module.initialize(outputscale=outputscale)
    likelihood.initialize(noise=noise)
    print("Initialized with: lengthscale=%.3f variance=%.3f noise=%.5f" % (model.covar_module.base_kernel.lengthscale.item(),
            model.covar_module.outputscale.item(),
            model.likelihood.noise.item()))

    model.train()
    likelihood.train()


    # Prediction
    # Get into evaluation (predictive posterior) mode
    model.eval()
    likelihood.eval()

    xtest = np.linspace(0, 1, 100)
    xtest2 = 0.5*np.ones([100,2]); xtest2[:,0] = xtest

    #%% Make predictions by feeding model through likelihood
    with torch.no_grad(), gpytorch.settings.fast_pred_var():
        predi = likelihood(model(torch.from_numpy(xtest2).float()))
        
        #plot(xtrain[:,0], ytrain, 'x')
        lower, upper = predi.confidence_region()
        yvar = predi.covariance_matrix.diag().numpy()
        plt.figure(figsize=(10,10))
        plt.plot(xtest2[:,0].flatten(), predi.mean.numpy())
        plt.fill_between(xtest2[:,0].flatten(), lower.numpy(), upper.numpy(), alpha=0.3)



# %%



from profit.sur import Surrogate
from profit.sur.gaussian_process import GPySurrogate
import numpy as np


Xtrain = np.array([1, 2, 3, 4]).reshape(-1, 1)
ytrain = np.array([5, 6, 7, 8]).reshape(-1, 1)


def test_default_hyperparameters():
    from profit.sur.gaussian_process import GPySurrogate

    sur = GPySurrogate()
    sur.prepare_train(Xtrain, ytrain)
    assert sur.hyperparameters['length_scale'] == [0.625]
    assert sur.hyperparameters['sigma_n'] == [0.03]
    assert sur.hyperparameters['sigma_f'] == np.std(ytrain, axis=0)


def test_add_training_data():
    Xadd = np.array([11, 12]).reshape(-1, 1)
    yadd = np.array([9, 10]).reshape(-1, 1)
    Xfull = np.concatenate([Xtrain, Xadd])
    yfull = np.concatenate([ytrain, yadd])
    for label in ('Custom', 'GPy', 'Sklearn'):
        sur = Surrogate[label]()
        sur.Xtrain = Xtrain
        sur.ytrain = ytrain
        if label == 'GPy':
            from GPy.models import GPRegression
            sur.model = GPRegression(Xtrain, ytrain)
        sur.add_training_data(Xadd, yadd)
        assert np.all(sur.Xtrain == Xfull)
        assert np.all(sur.ytrain == yfull)
        if label == 'GPy':
            assert np.all(sur.model.X == Xfull)
            assert np.all(sur.model.Y == yfull)


def test_select_kernel():
    from profit.sur.backend.python_kernels import RBF as cRBF
    from GPy.kern import RBF as gRBF, Matern52 as gMatern52
    sur = Surrogate['Custom']()
    assert sur.select_kernel('RBF') == cRBF

    sur = Surrogate['GPy']()
    sur.ndim = 2
    assert type(sur.select_kernel('RBF')) == gRBF
    assert all(hasattr(sur.select_kernel('RBF*Matern52'), subkernel) for subkernel in ('rbf', 'Mat52'))


def test_set_hyperparameters():
    from GPy.models import GPRegression
    from GPy.kern import RBF as gRBF, Matern52

    expected_hyperparameters = {'length_scale': np.array([1]), 'sigma_n': np.array([1]), 'sigma_f': np.array([1])}

    sur = GPySurrogate()

    # Default RBF kernel
    sur.model = GPRegression(Xtrain, ytrain)
    sur._set_hyperparameters_from_model()
    assert sur.hyperparameters == expected_hyperparameters

    # Product kernel
    sur.model = GPRegression(Xtrain, ytrain, kernel=gRBF(1)*Matern52(1))
    sur._set_hyperparameters_from_model()
    assert sur.hyperparameters == expected_hyperparameters


def test_default_Xpred():
    from profit.sur.gaussian_process import GPySurrogate

    # Instantiate surrogate and set relevant attributes.
    sur = GPySurrogate()
    sur.Xtrain = Xtrain
    sur.ndim = 1
    Xpred = sur.default_Xpred()
    assert np.all(Xpred == np.linspace(Xtrain[0], Xtrain[-1], 50).reshape(-1, 1))



from numpy import cos, linspace
from matplotlib.pyplot import plot, show
from profit.util import quasirand
from profit.sur.gaussian_process import GPySurrogate


# Original model
def f(u): return u + cos(10*u)


# Training response surface model
# on space-filling quasi-random set of points
xtrain = quasirand(npoint=10, ndim=1)
ytrain = f(xtrain)  # original response
fresp = GPySurrogate()
fresp.train(xtrain, ytrain, kernel='Matern52')  # fit profit model

# Evaluating response surface model
xtest  = linspace(0, 1, 100).reshape(-1, 1)  # points where to test
y, yvar = fresp.predict(xtest)  # prediction and variance

# Plot reference and response model fit
plot(xtest, f(xtest), color='k')
fresp.plot(xtest)
show()



"""High-dimensional output

This script drafts the following use-case: we make a parameter study over some
input parameters x and the domain code yields an output vector y with many
parameters. This is typically the case, when function-valued or even
"pixel-valued" output is produced, like on a 2D map (like 1024x1024).
Before fitting an input-output relation, one has to reduce the output to
a manageable number of dimensions. This means extracting the relevant
features.

Imagine the following idea: instead of viewing the output as many 1D outputs,
we look at it as a single vector in a high-dimensional space. Linear
combinations between vectors give new vectors. We can choose y from our
training data to span a basis in this vector space.

"""
#%%
import numpy as np
import matplotlib.pyplot as plt
from scipy.sparse.linalg.eigen import eigsh

def f(u, x):
    return np.cos(u*x)

x = np.linspace(0, 20, 100)

ntrain = 50

utrain = np.random.rand(ntrain)
plt.figure()

ytrain = []
for u in utrain:
    ytrain.append(f(u, x))
    plt.plot(x, ytrain[-1])
plt.xlabel('Independent variable x')
plt.ylabel(f'Output f(x;u) for different u')
ytrain = np.array(ytrain)

"""
Now we want to write
y_4 = a_1 y_1 + a_2 y_2 + a_3 y_3

This can be done by projection with inner products:

y_1*y_4 = a_1 y_1*y_1 + a_2 y_1*y_2 + a_3 y_1*y_3
y_2*y_4 = a_1 y_2*y_1 + a_2 y_2*y_2 + a_3 y_2*y_3
y_3*y_4 = a_1 y_3*y_1 + a_2 y_3*y_2 + a_3 y_3*y_3

We see that we have to solve a linear system with the
covariance matrix M_ij = y_i*y_j . To find the most relevant features
and reduce dimensionality, we use a PCA with only the highest eigenvalues.
We center the covariance around the mean, i.e. subtract ymean before.
"""
ymean = np.mean(ytrain, 0)
dytrain = ytrain - ymean
M = dytrain @ dytrain.T

w, Q = eigsh(M, 10)
plt.figure()
plt.plot(w)

plt.figure()
for i in range(len(w)):
    plt.plot(w[i]*Q[:,i] @ dytrain)
plt.title('10 most significant eigenvectors')
plt.xlabel('Independent variable x')
plt.ylabel('Eigenvectors scaled by eigenvalue')
#%% Testing

utest = 0.7
ytest = f(utest, x)
b = np.empty(ntrain)
for i in range(ntrain):
    b[i] = (ytest - ymean) @ dytrain[i]

a = Q @ (np.diag(1.0/w) @ (Q.T @ b))  # Brackets to avoid production of matrices
#a = np.linalg.solve(M, b)

plt.figure()
plt.plot(ymean + a @ dytrain, '--')
plt.plot(ytest)
plt.xlabel('Independent variable x')
plt.ylabel(f'Output f(x;u={utest})')
plt.legend(['Reconstruction', 'Reference'])

#%% Test with SVD
U, S, Vt = np.linalg.svd(dytrain, full_matrices=False)

# %%



import os
from shutil import copytree, rmtree, ignore_patterns
from profit import util


def rec2dict(rec):
    return {name: rec[name] for name in rec.dtype.names}


def write_input(eval_points, filename='input.txt'):
    """ Create input file with parameter combinations. """
    util.save(eval_points, filename)


def fill_run_dir(eval_points, template_dir='template/', run_dir='run/', param_files=None, overwrite=False):
    """ Fill each run directory with input data according to template format. """
    from tqdm import tqdm

    kruns = tqdm(range(eval_points.size))  # run with progress bar

    for krun in kruns:

        # .zfill(3) is an option that forces krun to have 3 digits
        run_dir_single = os.path.join(run_dir, str(krun).zfill(3))
        fill_run_dir_single(eval_points[krun], template_dir, run_dir_single, param_files, overwrite)


def fill_run_dir_single(params, template_dir, run_dir_single, param_files=None, overwrite=False,
                        ignore_path_exists=False):
    if os.path.exists(run_dir_single) and not ignore_path_exists:  # ToDo: make ignore_path_exists default
        if overwrite:
            rmtree(run_dir_single)
        else:
            raise RuntimeError('Run directory not empty: {}'.format(run_dir_single))
    copy_template(template_dir, run_dir_single)

    fill_template(run_dir_single, params, param_files=param_files)


def copy_template(template_dir, out_dir, dont_copy=None):
    """ TODO: explain dont_copy patterns """

    if dont_copy:
        copytree(template_dir, out_dir, symlinks=True, ignore=ignore_patterns(*dont_copy))
    else:
        copytree(template_dir, out_dir, symlinks=True)
    convert_relative_symlinks(template_dir, out_dir)


def convert_relative_symlinks(template_dir, out_dir):
    """ When copying the template directory to the single run directories,
     relative paths in symbolic links are converted to absolute paths. """
    for root, dirs, files in os.walk(out_dir):
        for filename in files:
            filepath = os.path.join(root, filename)
            if os.path.islink(filepath):
                linkto = os.readlink(filepath)
                if linkto.startswith('.'):
                    os.remove(filepath)
                    start_dir = os.path.relpath(root, out_dir)
                    os.symlink(os.path.join(template_dir, start_dir, filename), filepath)


def fill_template(out_dir, params, param_files=None):
    """
    Arguments:
        param_files(list): a list of filenames which are to be substituted or None for all
    """
    if param_files is None:
        param_files = []
    for root, dirs, files in os.walk(out_dir):  # by default, walk ignores subdirectories which are links
        for filename in files:
            filepath = os.path.join(root, filename)
            if (not param_files and not os.path.islink(filepath)) or filename in param_files:
                fill_template_file(filepath, filepath, params)


def fill_template_file(template_filepath, output_filepath, params, copy_link=True):
    """ Fill template in `template_filepath` by `params` and output into
    `output_filepath`. If `copy_link` is set (default), do not write into
    symbolic links but copy them instead.
    """
    with open(template_filepath, 'r') as f:
        content = replace_template(f.read(), params)
    if copy_link and os.path.islink(output_filepath):
        os.remove(output_filepath)  # otherwise the link target would be substituted
    with open(output_filepath, 'w') as f:
        f.write(content)


def replace_template(content, params):
    """ Returns filled template by putting values of `params` in `content`."""
    # Escape '{*}' for e.g. json templates by replacing it with '{{*}}'.
    # Variables then have to be declared as '{{*}}' which is replaced by a single '{*}'.
    pre, post = '{', '}'
    if '{{' in content:
        content = content.replace('{{', '§').replace('}}', '§§') \
            .replace('{', '{{').replace('}', '}}').replace('§§', '}').replace('§', '{')
        pre, post = '{{', '}}'
    return content.format_map(util.SafeDict.from_params(params, pre=pre, post=post))


def get_eval_points(config):
    """ Create input data as numpy array from config information.
    Use corresponding variable kinds (e.g. Uniform, Normal, Independent, etc.)
    """

    import numpy as np

    inputs = config['input']

    npoints = config['ntrain']
    dtypes = [(key, inputs[key]['dtype']) for key in inputs.keys()]

    eval_points = np.zeros((npoints, 1), dtype=dtypes)

    for n, (k, v) in enumerate(inputs.items()):
        eval_points[k] = np.round(v['range']) if np.issubdtype(eval_points[k].dtype, np.integer) else v['range']

    return eval_points



import numpy as np


class Encoder:
    r"""Class to handle encoding and decoding of the input and output data before creating the surrogate model.

    Work in progress!
    TODO: This should be implemented thoroughly in v0.5.

    Parameters:
        function_str (str): label for the encoding function. Currently only 'log10' and 'normalization'
            is supported.
        columns (list of int): Dimensions of the data the encoder acts on.
        output (bool): True if the encoder is for output data, False if the encoder works on input data.

    Attributes:
        label (str): Label for the encoding function.
        encode_func (function): Function used for encoding the data. E.g. $\log_{10}$.
        decode_func (function): Inverse transform of the encoding function. For an encoding of $\log_{10}(x)$ this
            would be $10^x$.
        variables (dict): Miscellaneous variables stored during encoding, which are needed for decoding. E.g. the
            scaling factor during normalization.
    """

    def __init__(self, function_str, columns, output=False):
        self.label = function_str
        self.encode_func, self.decode_func = self.get_function_from_str()
        self.variables = {}
        self.columns = columns
        self.output = output

        self.repr = [self.label, self.columns, self.output]

    def encode(self, x):
        """Applies the encoding function on given columns.

        Parameters:
            x (ndarray): Array to which the encoding is applied.

        Returns:
            ndarray: A copy of the encoded array x.
        """
        _x = x.copy()
        _x[:, self.columns] = self.encode_func(_x[:, self.columns])
        return _x

    def decode(self, x):
        """Applies the decoding function on given columns.

        Parameters:
            x (ndarray): Array to which the decoding is applied.

        Returns:
            ndarray: A copy of the decoded array x.
        """
        _x = x.copy()
        _x[:, self.columns] = self.decode_func(_x[:, self.columns])
        return _x

    def get_function_from_str(self):
        """Returns the encoding and decoding functions according to the identifier label.

        Returns:
            tuple: a tuple containing:
                - encoding_func (function): The function to encode data.
                - decoding_func (function): The inverse of the encoding function to decode the data again.
        """
        if self.label.lower() == "log10":
            return np.log10, lambda x: 10**x

        elif self.label.lower() == "normalization":
            def normalization(x):
                self.variables['xmax'] = abs(x).max(axis=0)
                return x / self.variables['xmax']

            def scaling(x):
                return x * self.variables['xmax']

            return normalization, scaling

        else:
            raise NotImplementedError("The encoding {} is not implemented yet.".format(self.label))



""" zeromq Interface

Ideas & Help from the 0MQ Guide (zguide.zeromq.org, examples are licensed with MIT)
"""

from .runner import RunnerInterface
from .worker import Interface

import zmq
import numpy as np
import json
from time import sleep
from logging import Logger
import os


@RunnerInterface.register('zeromq')
class ZeroMQRunnerInterface(RunnerInterface):
    """ Runner-Worker Interface using the lightweight message queue `ZeroMQ <https://zeromq.org/>`_

    - can use different transport systems, most commonly tcp
    - can be used efficiently on a cluster (tested)
    - expected to be inefficient for a large number of small, locally run simulations where communication overhead is
      a concern (unverified, could be mitigated by using a different transport system)
    - known issue: some workers were unable to establish a connection with three tries, reason unknown
    """
    def __init__(self, config, size, input_config, output_config, *, logger_parent: Logger = None):
        if 'FLAGS' not in [var[0] for var in self.internal_vars]:
            self.internal_vars += [('FLAGS', np.byte)]
        super().__init__(config, size, input_config, output_config, logger_parent=logger_parent)
        self.socket = zmq.Context.instance().socket(zmq.ROUTER)
        if self.config['address'] is None:
            bind = f'{self.config["transport"]}://*:{self.config["port"]}'
        else:
            bind = self.config['bind']
        self.socket.bind(bind)
        self.logger.info(f'connected to {bind}')

    def poll(self):
        self.logger.debug('polling: checking for messages')
        while self.socket.poll(timeout=10, flags=zmq.POLLIN):
            msg = self.socket.recv_multipart()
            # ToDo: Heartbeats
            self.handle_msg(msg[0], msg[2:])

    def handle_msg(self, address: bytes, msg: list):
        if address[:4] == b'req_':  # req_123
            run_id = int(address[4:])
            self.logger.debug(f'received {msg[0]} from run {run_id}')
            if msg[0] == b'READY':
                input_descr = json.dumps(self.input_vars).encode()
                output_descr = json.dumps(self.output_vars).encode()
                self.logger.debug(f'send input {input_descr} + {self.input[run_id]} + output {output_descr}')
                self.socket.send_multipart([address, b'', input_descr, self.input[run_id], output_descr])
                self.internal['FLAGS'][run_id] |= 0x02
            elif msg[0] == b'DATA':
                self.output[run_id] = np.frombuffer(msg[1], dtype=self.output_vars)
                self.logger.debug(f'received output {np.frombuffer(msg[1], dtype=self.output_vars)}')
                self.internal['DONE'][run_id] = True
                self.internal['FLAGS'][run_id] |= 0x08
                self.logger.debug('acknowledge DATA')
                self.socket.send_multipart([address, b'', b'ACK'])  # acknowledge
            elif msg[0] == b'TIME':
                self.internal['TIME'][run_id] = np.frombuffer(msg[1], dtype=np.uint)
                self.logger.debug('acknowledge TIME')
                self.socket.send_multipart([address, b'', b'ACK'])  # acknowledge
            elif msg[0] == b'DIE':
                self.internal['FLAGS'][run_id] |= 0x04
                self.logger.debug('acknowledge DIE')
                self.socket.send_multipart([address, b'', b'ACK'])  # acknowledge
            else:
                self.logger.warning(f'received unknown message {address}: {msg}')
        else:
            self.logger.warning(f'received message from unknown client {address}: {msg}')

    def clean(self):
        self.logger.debug('cleaning: closing socket')
        self.socket.close(0)

    @classmethod
    def handle_config(cls, config, base_config):
        """
        Example:
            .. code-block:: yaml

                class: zeromq
                transport: tcp      # transport system used by zeromq
                port: 9000          # port for the interface
                address: null       # override bind address used by zeromq
                connect: null       # override connect address used by zeromq
                timeout: 2500       # zeromq polling timeout, in ms
                retries: 3          # number of zeromq connection retries
                retry-sleep: 1      # sleep between retries, in s

        """
        defaults = dict(transport='tcp', port=9000, address=None, connect=None, timeout=2500, retries=3)
        defaults['retry-sleep'] = 1
        for key, value in defaults.items():
            if key not in config:
                config[key] = value


@Interface.register('zeromq')
class ZeroMQInterface(Interface):
    """ Runner-Worker Interface using the lightweight message queue `ZeroMQ <https://zeromq.org/>`_

    counterpart to :py:class:`ZeroMQRunnerInterface`
    """
    def __init__(self, config, run_id: int, *, logger_parent: Logger = None):
        super().__init__(config, run_id, logger_parent=logger_parent)
        self.connect()  # self.socket
        self._done = False
        self._time = 0
        self.request('READY')  # self.input, self.output

    @property
    def time(self):
        return self._time

    @time.setter
    def time(self, value: int):
        self._time = value
        self.request('TIME')

    def done(self):
        self.request('DATA')
        self.socket.close(0)

    def connect(self):
        self.socket = zmq.Context.instance().socket(zmq.REQ)
        self.socket.setsockopt(zmq.IDENTITY, f'req_{self.run_id}'.encode())
        if self.config['connect'] is None:
            address = os.environ.get('PROFIT_RUNNER_ADDRESS')
            if address is None:
                address = 'localhost'
            connect = f'{self.config["transport"]}://{address}:{self.config["port"]}'
        else:
            connect = self.config['connect']
        self.socket.connect(connect)
        self.logger.info(f'connected to {connect}')

    def request(self, request):
        """ 0MQ - Lazy Pirate Pattern """
        if request not in ['READY', 'DATA', 'TIME']:
            raise ValueError('unknown request')
        tries = 0
        while True:
            msg = [request.encode()]
            if request == 'DATA':
                msg.append(self.output)
            elif request == 'TIME':
                msg.append(np.uint(self.time))
            self.socket.send_multipart(msg)
            if self.socket.poll(timeout=self.config['timeout'], flags=zmq.POLLIN):
                response = None
                try:
                    response = self.socket.recv_multipart()
                    if request == 'READY':
                        input_descr, input_data, output_descr = response
                        input_descr = [tuple(column) for column in json.loads(input_descr.decode())]
                        output_descr = [tuple(column[:2] + [tuple(column[2])]) for column
                                        in json.loads(output_descr.decode())]
                        self.input = np.frombuffer(input_data, dtype=input_descr)[0]
                        self.output = np.zeros(1, dtype=output_descr)[0]
                        self.logger.info('READY: received input data')
                        self.logger.debug(f'received: {np.frombuffer(input_data, dtype=input_descr)}')
                        return
                    else:
                        assert response[0] == b'ACK'
                        self.logger.debug(f'{request}: message acknowledged')
                        return
                except (ValueError, AssertionError):
                    self.logger.debug(f'{request}: received {response}')
                    self.logger.error(f'{request}: malformed reply')
            else:
                self.logger.warning(f'{request}: no response')
                tries += 1
                sleep(self.config['retry-sleep'])

            if tries >= self.config['retries'] + 1:
                self.logger.error(f'{request}: {tries} requests unsuccessful, abandoning')
                self.socket.close(0)
                raise ConnectionError('could not connect to RunnerInterface')

            # close and reopen the socket
            self.socket.close(linger=0)
            self.connect()



# Configuration file for RedMod scripts v0.1
from redmod import config, uq

# Backend for uncertainty quantification
uq.backend = uq.ChaosPy(order = 3, sparse = True)

# Parameters for uncertainty quantification
#uq.params['u'] = uq.Normal(5.0, 0.3)
uq.params['u'] = uq.Uniform(4.7, 5.3)
uq.params['v'] = uq.Uniform(0.55, 0.6)

# How to run the model code
config.command = 'python mockup.py'



""" Scheduling runs on a cluster with SLURM

* targeted towards aCluster@tugraz.at
* each run is submitted as a job using a slurm batch script
* run arrays are submitted as a slurm job array
* by default completed runs are recognised by the interface, but the scheduler is polled as a fallback (less often)
"""

from .runner import Runner

import subprocess
from time import sleep, time
import os
from tqdm import trange


@Runner.register('slurm')
class SlurmRunner(Runner):
    """ Runner which submits each run as a job to the SLURM scheduler on a cluster

    - generates a slurm batch script with the given configuration
    - can also be used with a custom script
    - supports OpenMP
    - tries to minimize overhead by using job arrays if possible
    - polls the scheduler only at longer intervals
    """
    def __init__(self, interface_class, config, base_config):
        super().__init__(interface_class, config, base_config)
        if self.config['custom']:
            if not os.path.exists(self.config['path']):
                self.logger.error(f'flag for custom script is set, but could not be found at '
                                  f'specified location {self.config["path"]}')
                self.logger.debug(f'cwd = {os.getcwd()}')
                self.logger.debug(f'ls = {os.listdir(os.path.dirname(self.config["path"]))}')
                raise FileNotFoundError(f'could not find {self.config["path"]}')
        else:
            self.generate_script()

    def spawn_run(self, params=None, wait=False):
        super().spawn_run(params, wait)  # fill data with params
        self.logger.info(f'schedule run {self.next_run_id:03d} via Slurm')
        self.logger.debug(f'wait = {wait}, params = {params}')
        env = self.env.copy()
        env['PROFIT_RUN_ID'] = str(self.next_run_id)
        submit = subprocess.run(['sbatch', '--parsable', self.config['path']],
                                cwd=self.base_config['run_dir'], env=env, capture_output=True, text=True, check=True)
        job_id = submit.stdout.split(';')[0].strip()
        self.runs[self.next_run_id] = job_id
        if wait:
            self.wait_for(self.next_run_id)
        self.next_run_id += 1

    def spawn_array(self, params_array, blocking=True):
        self.logger.info(f'schedule array {self.next_run_id} - {self.next_run_id + len(params_array) - 1} via slurm')
        self.fill(params_array, offset=self.next_run_id)
        env = self.env.copy()
        env['PROFIT_RUN_ID'] = str(self.next_run_id)
        array_str = f'--array=0-{len(params_array) - 1}'
        if self.config['parallel'] is not None:
            array_str += f'%{self.config["parallel"]}'
        submit = subprocess.run(['sbatch', '--parsable', array_str, self.config['path']],
                                cwd=self.base_config['run_dir'], env=env, capture_output=True, text=True, check=True)
        job_id = submit.stdout.split(';')[0].strip()
        for i in range(len(params_array)):
            self.runs[self.next_run_id + i] = f'{job_id}_{i}'
        if blocking:
            self.wait_for_all([self.next_run_id + i for i in range(len(params_array))], show_tqdm=True)
        self.next_run_id += len(params_array)

    def wait_for_all(self, run_ids, show_tqdm=False):
        """wait until all specified runs have completed"""
        poll_time = time()
        if show_tqdm:
            progress = trange(len(run_ids))
        while len([i for i in run_ids if i in self.runs]):
            self.check_runs()
            sleep(self.config['sleep'])
            # poll the scheduler after a longer period
            if time() - poll_time > self.config['poll']:
                self.check_runs(poll=True)
                poll_time = time()
                sleep(self.config['sleep'])
            if show_tqdm:
                progress.update(len([i for i in run_ids if i not in self.runs]) - progress.n)

    def wait_for(self, run_id: int):
        """wait until the specified run has completed"""
        self.wait_for_all([run_id])

    def check_runs(self, poll: bool = False):
        """check the status of runs via the interface, poll only when specified"""
        # ask interface and remove all completed runs
        self.interface.poll()
        for run_id in list(self.runs):
            if self.interface.internal['DONE'][run_id]:
                self.del_run(run_id)
        # poll the slurm scheduler to check for crashed runs
        if poll:
            acct = subprocess.run(['sacct', f'--name={self.config["options"]["job-name"]}', '--brief', '--parsable2'],
                                  capture_output=True, text=True, check=True)
            lookup = {job: run for run, job in self.runs.items()}
            for line in acct.stdout.split('\n'):
                if len(line) < 2:
                    continue
                job_id, state = line.split('|')[:2]
                if job_id in lookup:
                    if not (state.startswith('RUNNING') or state.startswith('PENDING')):
                        self.del_run(lookup[job_id])

    def cancel_all(self):
        from re import split
        ids = set()
        for run_id in self.runs:
            ids.add(split(r'[_.]', self.runs[run_id])[0])
        for job_id in ids:
            subprocess.run(['scancel', job_id])
        self.runs = {}

    def del_run(self, run_id: int):
        """helper: delete run from runs and remove slurm-stdout

        :param run_id: which run to delete, is a key in ``self.runs``
        """
        if self.run_config['clean'] and self.interface.internal['DONE'][run_id]:
            path = os.path.join(self.base_config['run_dir'], f'slurm-{self.runs[run_id]}.out')
            try:
                os.remove(path)
            except FileNotFoundError:
                pass
        del self.runs[run_id]

    def clean(self):
        """remove generated scripts and any slurm-stdout-files which match ``slurm-*.out``"""
        super().clean()
        if not self.config['custom'] and os.path.exists(self.config['path']):
            os.remove(self.config['path'])
        for direntry in os.scandir(self.base_config['run_dir']):
            if direntry.is_file() and direntry.name.startswith('slurm-') and direntry.name.endswith('.out'):
                os.remove(os.path.join(self.base_config['run_dir'], direntry.path))

    @classmethod
    def handle_config(cls, config, base_config):
        """
        Example:
            .. code-block:: yaml

                class: slurm
                parallel: null      # maximum number of simultaneous runs (for spawn array)
                sleep: 0            # number of seconds to sleep while (internally) polling
                poll: 60            # number of seconds between external polls (to catch failed runs), use with care!
                path: slurm.bash    # the path to the generated batch script (relative to the base directory)
                custom: false       # whether a custom batch script is already provided at 'path'
                prefix: srun        # prefix for the command
                OpenMP: false       # whether to set OMP_NUM_THREADS and OMP_PLACES
                cpus: 1             # number of cpus (including hardware threads) to use (may specify 'all')
                options:            # (long) options to be passed to slurm: e.g. time, mem-per-cpu, account, constraint
                    job-name: profit

        """
        defaults = {'parallel': None,
                    'sleep': 0,
                    'poll': 60,
                    'path': 'slurm.bash',
                    'custom': False,
                    'prefix': 'srun',
                    'OpenMP': False,
                    'cpus': 1,
                    }
        options = {'job-name': 'profit'}

        for key, value in defaults.items():
            if key not in config:
                config[key] = value
        if 'options' not in config:
            config['options'] = {}
        for key, value in options.items():
            if key not in config['options']:
                config['options'][key] = value

        # convert path to absolute path
        if not os.path.isabs(config['path']):
            config['path'] = os.path.abspath(os.path.join(base_config['base_dir'], config['path']))
        # check type of 'cpus'
        if (type(config['cpus']) is not int or config['cpus'] < 1) and config['cpus'] != 'all':
            raise ValueError(f'config option "cpus" may only be a positive integer or "all" and not {config["cpus"]}')

    def generate_script(self):
        text = f"""\
#!/bin/bash
# automatically generated SLURM batch script for running simulations with proFit
# see https://github.com/redmod-team/profit
"""
        for key, value in self.config['options'].items():
            if value is not None:
                text += f"\n#SBATCH --{key}={value}"

        text += """
#SBATCH --ntasks=1
"""
        if self.config['cpus'] == 'all':
            text += """
#SBATCH --nodes=1
#SBATCH --exclusive"""
        elif self.config['cpus'] > 1:
            text += f"""
#SBATCH --cpus-per-task={self.config['cpus']}"""

        if self.config['OpenMP']:
            text += """
export OMP_NUM_THREADS=$SLURM_CPUS_ON_NODE
export OMP_PLACES=threads"""

        text += f"""
if [[ -n $SLURM_ARRAY_TASK_ID ]]; then
    export PROFIT_ARRAY_ID=$SLURM_ARRAY_TASK_ID
fi
export PROFIT_RUNNER_ADDRESS=$SLURM_SUBMIT_HOST

{self.config["prefix"]} {'profit-worker' if not self.run_config['custom'] else self.run_config['command']}
"""
        with open(self.config['path'], 'w') as file:
            file.write(text)



#%%
import numpy as np
import GPy

k0 = GPy.kern.RBF(1, active_dims=0)
k1 = GPy.kern.RBF(1, active_dims=1)

k0_der = GPy.kern.DiffKern(k0, 0)

# Extended class for product kernel,
# can be merged with Prod class when finished
class ProdExtended(GPy.kern.Prod):
    def dK_dX(self, X, X2, dimX):
        raise(NotImplementedError)

    def dK2_dXdX2(self, X, X2, dimX, dimX2):
        raise(NotImplementedError)

kprod = ProdExtended((k0, k1))

x0train = np.array([0.0, 1.0, 0.0]).reshape(-1,1)
x1train = np.array([0.0, 0.0, 1.0]).reshape(-1,1)
xtrain = np.hstack((x0train, x1train))


print('Training points:')
print(xtrain)
print()

print('K0 = ')
print(k0.K(x0train, x0train))
print(k0.dK_dX(x0train, x0train, 0))
print(k0.dK2_dXdX2(x0train, x0train, 0, 0))
print()

print('K1 = ')
print(k1.K(xtrain, xtrain))  # Need 2D vectors, here, as active_dims=1
print()

print('Prod K = ')
print(kprod.K(xtrain, xtrain))
print()

#%% TODO: test 1st and 2nd derivatives based on parts of Prod kernel
X = xtrain
X2 = xtrain
dimX = 0
# Apply product rule to kprod = k0*k1
# so    d(kprod/dx0) = dk0/dx0*k1
# and   d(kprod/dx1) = k0*dk1/dx1

other = kprod.parts[:]      # to store all parts except the dimX one
diffpart = other.pop(dimX)  # removes dimX and returns it as diffpart

print('Part to differentiate:')
print(diffpart)  # should give k0 here
print('Other parts')
print(other[0])  # should give k1 here

dK_dX_diffpart = diffpart.dK_dX(X, X2, dimX)
K_other = [k.K(X, X2) for k in other]
result = dK_dX_diffpart*np.prod(K_other)

print('Derivative factor:')
print(dK_dX_diffpart)
print('Other factors:')
print(K_other)
print('Overall result:')
print(result)

#%% This is not implemented yet, TODO in class
print('Derivatives Prod K')
print(kprod.dK_dX(xtrain, xtrain, 0))
print(kprod.dK2_dXdX2(xtrain, xtrain, 0, 0))
print()

#%%